/******************************************************************************
 *
 * Driver for Option High Speed Mobile Devices.
 *
 *  Copyright (C) 2008  Option International
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *	0.1	Option International
 *		2.4 driver 
 *
 *****************************************************************************/

/******************************************************************************
 *
 * Description of the device:
 *
 * Interface 0:	Contains the IP network interface on the bulk end points.
 *		The multiplexed serial ports are using the interrupt and control endpoints.
 *		Interrupt contains a bitmap telling which multiplexed serialport needs servicing.
 *
 * Interface 1:	Diagnostics port, uses bulk only, do not submit urbs until the port is opened, as
 *		this have a huge impact on the network port throughput.
 *
 * Interface 2:	Standard modem interface - circuit switched interface, should not be used.
 *
 *****************************************************************************/


#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/etherdevice.h>
#include <linux/module.h>
#include <linux/ethtool.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/kmod.h>
#include <asm/byteorder.h>
#include <linux/proc_fs.h>
#include <net/arp.h>
#include <linux/ip.h>

//*****************************************************************************
// Defines
//*****************************************************************************
// #define DEBUG

#define DRIVER_VERSION			"0.1"
#define MOD_AUTHOR			"Option Wireless"
#define MOD_DESCRIPTION			"USB High Speed Option driver"
#define MOD_LICENSE			"GPL"

/* From CDC Spec Table 24 */
#define CS_INTERFACE_DESC		0x24

#define HSO__MAX_MTU			1984
#define DEFAULT_MTU			1500
#define DEFAULT_MRU			1500

#define CTRL_URB_RX_SIZE		1024  
#define CTRL_URB_TX_SIZE		64

#define BULK_URB_RX_SIZE		4096
#define BULK_URB_TX_SIZE		8192

#define MUX_INTR_BUF_SIZE		16
#define MUX_BULK_RX_BUF_SIZE		HSO__MAX_MTU
#define MUX_BULK_TX_BUF_SIZE		HSO__MAX_MTU
#define MUX_BULK_RX_BUF_COUNT		1
#define USB_TYPE_OPTION_VENDOR		0x20

/* These definitions are used with the hso_priv flags element - use *_bit operations on it. (bit indices not values.)*/
#define HSO_NET_PRESENT			0	/* 0x00000001 */
#define HSO_NET_RUNNING			1	/* 0x00000002 */
#define HSO_NET_TX_BUSY			2	/* 0x00000004 */
#define HSO_NET_RX_BUSY			3	/* 0x00000008 */
#define HSO_NET_UNPLUG			4	/* 0x00000010 */

#define	HSO_NET_TX_TIMEOUT		(HZ*10)

#define SEND_ENCAPSULATED_COMMAND	0x00
#define GET_ENCAPSULATED_RESPONSE	0x01

/* Serial port defines and structs. */
#define HSO_RX_BUFFER_SIZE		8192
#define HSO_THRESHOLD_THROTTLE		(7*1024)
#define HSO_THRESHOLD_UNTHROTTLE	(2*1024)
#define HSO_RX_MAX_FLIP_SIZE		2048


/* These definitions used in the Ethernet Packet Filtering requests */
/* See CDC Spec Table 62 */
#define MODE_FLAG_PROMISCUOUS		(1<<0)
#define MODE_FLAG_ALL_MULTICAST		(1<<1)
#define MODE_FLAG_DIRECTED		(1<<2)
#define MODE_FLAG_BROADCAST		(1<<3)
#define MODE_FLAG_MULTICAST		(1<<4)

/* CDC Spec class requests - CDC Spec Table 46 */
#define SET_ETHERNET_MULTICAST_FILTER	0x40
#define SET_ETHERNET_PACKET_FILTER	0x43

#define HSO_PRIV_FLAG_THROTTLED		0
#define HSO_PRIV_FLAG_TX_SENT		1
#define HSO_PRIV_FLAG_RX_SENT		2

#define HSO_SERIAL_TTY_MAJOR		245
#define HSO_SERIAL_MAGIC		0x48534f31

#define HSO_SERIAL_TTY_MINORS		4       /*  Number of ttys to handle (including diagnostics port) */
#define HSO_SERIAL_TTY_CTRL_DEVS	2       /*  Plain TTYs (no diagnostics interface) */

//*****************************************************************************
// Debug
//*****************************************************************************
#define D__(lvl_, fmt, arg...)	do { printk(lvl_ "[%d:%s]: " fmt "\n", __LINE__, __FUNCTION__, ## arg); } while(0)

#define NFO(args...)		D__( KERN_INFO, ##args)
#define ERR(args...)		D__( KERN_ERR,  ##args)
#define WARN(args...)		D__( KERN_WARNING,  ##args)

#define D_(lvl, args...)	do { if(lvl & debug) { NFO( args ); } } while(0)
#define D1(args...)		D_(0x01, ##args)
#define D2(args...)		D_(0x02, ##args)
#define D3(args...)		D_(0x04, ##args)
#define D4(args...)		D_(0x08, ##args)
#define D5(args...)		D_(0x10, ##args)
#define D6(args...)		D_(0x20, ##args)
#define D			D1

#ifdef DEBUG
static void dbg_dump(int line_count, const char *func_name, unsigned char *buf, unsigned int len)
{
	u8 i = 0;
	u8 j = 0;

	printk(KERN_INFO "[%d:%s]: len %d\n" KERN_INFO, line_count, func_name, len);

	for(i=0; i<len; i++)
	{
		printk("\n" KERN_INFO "  0x%03x: ", i);
		for(j = i; j - i < 8 && j < len; j++)
		{
			printk("%02x ", (unsigned char)buf[j]);
		}
		printk(" ");
		for(j = i; j - i < 8 && j < len; j++)
		{
			if(buf[j] == '\n')
				printk("\\n");
			else if(buf[j] == '\r')
				printk("\\r");
			else
				printk("%c", (unsigned char)buf[j]);
		}
		i += j;
	}
	printk("\n");
}
#define DUMP(buf_, len_)	dbg_dump(__LINE__, __FUNCTION__, buf_, len_)
#define DUMP1(buf_, len_)	do{ if(0x01 & debug) DUMP(buf_, len_);}while(0)
#else	// DEBUG
#define DUMP(buf_, len_)
#define DUMP1(buf_, len_)
#endif	// DEBUG

//*****************************************************************************
// Type definitions
//*****************************************************************************

typedef enum {
	MUX_INTERFACE		= 0,
	QXDM_INTERFACE		= 1,
	CS_INTERFACE		= 2,
} type_intf;

typedef enum  {
	WAIT_IP,
	WAIT_DATA,
	WAIT_SYNC
} pkt_parse_state_t;

typedef struct {
	unsigned char *p_b, *p_h, *p_t;
	unsigned int cnt, size;
} fifo_t;

typedef struct {
	u8	length		__attribute__ ((packed));
	u8	descriptor_type	__attribute__ ((packed));
	u8	enabled_ports	__attribute__ ((packed));
} option_descriptor_t;

typedef struct _hso_priv {
	struct net_device_stats		stats;
	struct net_device		*net;
	struct usb_device		*usb;
	struct usb_interface		*interfaces[3];

	option_descriptor_t		option_info;
	u32				flags;
	u32				properties;
	int				configuration_num;
	int				bConfigurationValue;

	int				mux_bInterfaceNumber;
	int				mux_altsetting;

	int				mux_ep_intr;
	int				mux_ep_intr_size;
	u8				*mux_intr_buf;
	struct urb			*mux_intr_urb;

	int				mux_ep_bulk_in;
	int				mux_ep_bulk_in_size;
	int				mux_ep_bulk_out;
	int				mux_ep_bulk_out_size;
	int				mux_bInterval;
	struct urb			*mux_bulk_rx_urb_pool[MUX_BULK_RX_BUF_COUNT];
	struct urb			*mux_bulk_tx_urb;
	u8				*mux_bulk_rx_buf_pool[MUX_BULK_RX_BUF_COUNT];
	u8				*mux_bulk_tx_buf;
	struct sk_buff			*skb_rx_buf;
	pkt_parse_state_t		rx_parse_state;
	unsigned short			rx_buf_size;
	unsigned short			rx_buf_missing;
	struct iphdr			rx_ip_hdr;
	struct ethhdr			dummy_eth_head;

	__u16				bcdCDC;
	__u16				wMaxSegmentSize;
	__u16				wNumberMCFilters;
	__u16				mode_flags;

	int				transmitted;
} hso_priv;

typedef struct _hso_serial_t hso_serial_t;
struct _hso_serial_t {
	int				magic;
	u8                      	minor;          /* port index */
	int				type;
	/* rx/tx urb could be either a bulk urb or a control urb depending on which serial port it is used on. */
	struct urb*			rx_urb;
	u8*				rx_data;
	u16				rx_data_length; /* should contain allocated length */

	struct urb*			tx_urb;
	u8*				tx_data;
	u8*				tx_buffer;
	u16				tx_data_length; /* should contain allocated length */
	u16				tx_data_count;
	u16				tx_buffer_count;
	struct usb_ctrlrequest		ctrl_req_tx;
	struct usb_ctrlrequest		ctrl_req_rx;

	int				ep_bulk_in;
	int				ep_bulk_out;
	int				ep_bulk_out_size;
	int				ep_bulk_in_size;

	hso_priv			*odev;
	/* rx fifo handling from hss */
	fifo_t				rx_fifo;
	unsigned char			rx_buffer[HSO_RX_BUFFER_SIZE];
	struct tq_struct		rx_flip_task;
	int				flags;

	int (*write_data)(hso_serial_t *serial);

	struct tq_struct		tqueue;
	u8				read_scheduled;

	/* from usb_serial_port */
	struct tty_struct		*tty;
	int				open_count;
	spinlock_t			serial_lock;
	void				*private;
};


//*****************************************************************************
// Globals
//*****************************************************************************
static int			debug = 0x00;
static int			hotplug = 1;
static const char		driver_name[] = "hso";
static const char		*version = __FILE__ ": " DRIVER_VERSION " " MOD_AUTHOR;
static hso_priv			*global_odev = NULL;
static int			hso_serial_refcount;
static struct tty_struct	*hso_serial_tty[HSO_SERIAL_TTY_MINORS];
static struct termios		*hso_serial_termios[HSO_SERIAL_TTY_MINORS];
static struct termios		*hso_serial_termios_locked[HSO_SERIAL_TTY_MINORS];
static hso_serial_t		*hso_serial_table[HSO_SERIAL_TTY_MINORS] = {0};	/* initially all NULL */
static int			last_serial_minor_used = -1;
const unsigned char		dummy_mac[6] = {0,1,2,3,4,5};	/* TODO: make this a device variable, also used in arp_send */

//*****************************************************************************
// Prototypes
//*****************************************************************************
static int hso_net_open(struct net_device *net);
static int hso_net_close( struct net_device *net );
static int hso_net_start_xmit( struct sk_buff *skb, struct net_device *net );
static int hso_net_ioctl( struct net_device *net, struct ifreq *rq, int cmd );
static struct net_device_stats *hso_net_get_stats( struct net_device *net );
static void hso_net_tx_timeout( struct net_device *net );
static void hso_net_set_multicast( struct net_device *net );
static void read_bulk_callback( struct urb *urb );
static void packetizeRx(hso_priv *odev, unsigned char *ip_pkt, unsigned int count, unsigned char is_eop);
static void write_bulk_callback( struct urb *urb );
static int hso_handle_arp( struct sk_buff *skb, struct net_device *net ) ;
static void arp_send_response( struct net_device *dev, unsigned char *dst_hw, unsigned char *src_hw, u32 s_pa,u32 t_pa, unsigned char *s_ha, unsigned char *t_ha) ;
static int hso_serial_open(struct tty_struct *tty, struct file * filp);
static void hso_serial_close (struct tty_struct *tty, struct file * filp);
static int hso_serial_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count);
static int hso_serial_write_room (struct tty_struct *tty);
static void hso_serial_set_termios (struct tty_struct *tty, struct termios * old);
static int hso_serial_chars_in_buffer (struct tty_struct *tty);
static int hso_serial_ioctl (struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg);
static void hso_serial_throttle (struct tty_struct * tty);
static void hso_serial_unthrottle (struct tty_struct * tty);
static void hso_serial_break (struct tty_struct *tty, int break_state);
static int hso_serial_read_proc (char *page, char **start, off_t off, int count, int *eof, void *data);
static void hso_serial_hangup (struct tty_struct *tty);
static void hso_kick_transmit( hso_serial_t *serial );
static void hso_kick_receive( hso_serial_t *serial );
static void intr_callback( struct urb *urb );
static int hso_mux_serial_read(hso_priv *odev, hso_serial_t* serial);
static int hso_mux_serial_write_data(hso_serial_t *serial);
static int hso_std_serial_write_data(hso_serial_t *serial);
static int mux_device_request(hso_serial_t *serial, u8 type, u16 port, struct urb* ctrl_urb, struct usb_ctrlrequest *ctrl_req, u8* ctrl_urb_data, u32 size);
static void ctrl_callback( struct urb *urb );
static void put_rxbuf_data(struct urb *urb, hso_serial_t *serial);
static void hso_std_serial_read_bulk_callback( struct urb *urb );
static void hso_std_serial_write_bulk_callback( struct urb *urb ) ;
static void _hso_serial_set_termios (struct tty_struct *tty, struct termios * old);
int hso_read_info(char *buf, char **start, off_t offset, int count, int *eof, void *data);
int __init hso_init(void);
void __exit hso_exit(void);
static void* hso_probe( struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id);
static void hso_disconnect( struct usb_device *usb, void *ptr );
static int check_for_claimed_interfaces( struct usb_config_descriptor *config, hso_priv *odev );
static int get_mux_endpoints( struct usb_interface_descriptor *desc, hso_priv *odev);
static int hso_serial_init();
static void hso_serial_exit();
static hso_serial_t *hso_serial_start(hso_priv *odev, unsigned char minor, int type, unsigned char num);
static int get_std_serial_endpoints(hso_serial_t *serial, int ifacenr, hso_priv *odev);
static void hso_serial_disconnect(struct usb_device *usb, hso_priv *odev);
static void __hso_serial_close(hso_serial_t *serial);
static void hso_serial_stop(struct usb_device *dev, unsigned char minor, hso_priv *odev);
static int hso_alloc_memory(hso_priv **_odev);
static void hso_free_memory(hso_priv **_odev);
static void set_ethernet_addr( hso_priv *odev );
void log_device_info(hso_priv *odev);
static int parse_option_information( unsigned char *data, int length, hso_priv *odev );
static int find_valid_configuration( struct usb_device *device, hso_priv *odev );
static int check_interface( struct usb_device *device, hso_priv *odev, type_intf type );
static inline unsigned char hex2dec( unsigned char digit );
static void get_string(u8 *buf, int buf_len, int string_num, hso_priv *odev);
static void call_policy_interface (char *verb, int line, __u16 idVendor, __u16 idProduct);
static inline hso_serial_t *get_serial(struct tty_struct *tty);
static void hso_serial_rx_flip(void *priv);
static int get_rxbuf_data(unsigned char *data, unsigned short len, hso_serial_t *serial);
static void port_softint(void *private);

//*****************************************************************************
// Declarations
//*****************************************************************************
static struct usb_device_id hso_ids[] = {
	{USB_DEVICE(0x0af0, 0x6711)},
	{USB_DEVICE(0x0af0, 0x6731)},
	{USB_DEVICE(0x0af0, 0x6751)},
	{USB_DEVICE(0x0af0, 0x6771)},
	{USB_DEVICE(0x0af0, 0x6791)}, 
	{USB_DEVICE(0x0af0, 0x6811)},
	{USB_DEVICE(0x0af0, 0x6911)},
	{USB_DEVICE(0x0af0, 0x6951)},
	{USB_DEVICE(0x0af0, 0x6971)},
	{USB_DEVICE(0x0af0, 0x7011)},
	{USB_DEVICE(0x0af0, 0x7031)},
	{USB_DEVICE(0x0af0, 0x7051)},
	{USB_DEVICE(0x0af0, 0x7071)},
	{USB_DEVICE(0x0af0, 0x7111)},
	{USB_DEVICE(0x0af0, 0x7211)},
	{USB_DEVICE(0x0af0, 0x7251)},
	{USB_DEVICE(0x0af0, 0x7271)},
	{USB_DEVICE(0x0af0, 0x7311)},
	{USB_DEVICE(0x0af0, 0xc031)}, /* Icon-Edge */
	{USB_DEVICE(0x0af0, 0xd013)}, /* Module HSxPA */
	{USB_DEVICE(0x0af0, 0xd031)}, /* Icon-321 */
	{USB_DEVICE(0x0af0, 0xd033)}, /* Icon-322 */
	{ }
};

static struct usb_driver hso_driver = {
	name:		driver_name,
	probe:		hso_probe,
	disconnect:	hso_disconnect,
	id_table:	hso_ids,
};

static struct tty_driver hso_serial_tty_driver = {
	.magic =		TTY_DRIVER_MAGIC,
	.driver_name =		"usb-hso-serial",
#ifndef CONFIG_DEVFS_FS
	.name =			"ttyHS",
#else
	.name =			"usb/tth/%d",
#endif
	.major =		HSO_SERIAL_TTY_MAJOR,
	.minor_start =		0,
	.num =			HSO_SERIAL_TTY_MINORS,
	.type =			TTY_DRIVER_TYPE_SERIAL,
	.subtype =		SERIAL_TYPE_NORMAL,
	.flags =		TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS,

	.refcount =		&hso_serial_refcount,
	.table =		hso_serial_tty,
	.termios =		hso_serial_termios,
	.termios_locked =	hso_serial_termios_locked,

	.open =			hso_serial_open,
	.close =		hso_serial_close,
	.write =		hso_serial_write,
	.write_room =		hso_serial_write_room,
	.ioctl =		hso_serial_ioctl,
	.set_termios =		hso_serial_set_termios,
	.throttle =		hso_serial_throttle,
	.unthrottle =		hso_serial_unthrottle,
	.break_ctl =		hso_serial_break,
	.chars_in_buffer =	hso_serial_chars_in_buffer,
	.read_proc =		hso_serial_read_proc,
	.hangup =		hso_serial_hangup,
};


//*****************************************************************************
//*****************************************************************************
// Network interface functions
//*****************************************************************************
//*****************************************************************************

//=============================================================================
//	function:	hso_net_open
//	purpose:	called when net interface is brought up by ifconfig
//=============================================================================
static int hso_net_open(struct net_device *net)
{
	hso_priv *odev = (hso_priv *)net->priv;
	int result = 0;
	int i = 0;

	/* reset read counter. */
	odev->stats.rx_packets = 0;
	/* setup environment */
	odev->rx_parse_state = WAIT_IP;
	odev->rx_buf_size = 0;
	odev->rx_buf_missing = sizeof(struct iphdr);

	/* send URBs for all read buffers */
	for(i = 0; i < MUX_BULK_RX_BUF_COUNT; i++)
	{
		/* Prep a receive URB */
		usb_fill_bulk_urb(	odev->mux_bulk_rx_urb_pool[i],
					odev->usb,
					usb_rcvbulkpipe(odev->usb, odev->mux_ep_bulk_in),
					odev->mux_bulk_rx_buf_pool[i],
					MUX_BULK_RX_BUF_SIZE,
					read_bulk_callback,
					odev );

		odev->mux_bulk_rx_urb_pool[i]->transfer_flags |= USB_QUEUE_BULK;

		/* Put it out there so the device can send us stuff */
		if ( (result = usb_submit_urb(odev->mux_bulk_rx_urb_pool[i])) ) {
			WARN( "%s failed modem_bulk_rx_urb %d", __FUNCTION__, result );
		}
	}

	// Tell the kernel we are ready to start receiving from it
	netif_start_queue( net );

	// We are up and running.
	set_bit(HSO_NET_RUNNING, &odev->flags);

	// Let's get ready to move frames!!!
	return 0;
}

//=============================================================================
//	function:	hso_net_close
//	purpose:	called when interface is brought down by ifconfig
//=============================================================================
static int hso_net_close(struct net_device *net)
{
	hso_priv *odev = net->priv;
	int i = 0;

	/* no longer running */
	clear_bit(HSO_NET_RUNNING, &odev->flags);
	/* we don't need the queue anymore */
	netif_stop_queue(net);

	/* We don't need the net specific URBs anymore. (interrupt still needed for serial) */
	for(i = 0; i < MUX_BULK_RX_BUF_COUNT; i++) {
		usb_unlink_urb(odev->mux_bulk_rx_urb_pool[i]);
	}
	usb_unlink_urb( odev->mux_bulk_tx_urb );

	/* done */
	return 0;
}

//=============================================================================
//	function:	hso_net_start_xmit
//	purpose:	called by kernel when we need to transmit a packet
//=============================================================================
static int hso_net_start_xmit(struct sk_buff *skb, struct net_device *net)
{
	hso_priv *odev = net->priv;
	int result = 0;

	/* Check if it is an ARP packet. returns 1 if it is and has been dropped. */
	if(hso_handle_arp(skb, net)) {
		return 0;
	}

	/* Tell the kernel, "No more frames 'til we are done with this one." */
	netif_stop_queue(net);
	/* fetch the packet */
	skb_pull(skb, ETH_HLEN);
	/* log if asked */
	DUMP1(skb->data, skb->len);
	/* Copy it from kernel memory to OUR memory */
	memcpy(odev->mux_bulk_tx_buf, skb->data, skb->len);
	D1("len: %d/%d", skb->len, MUX_BULK_TX_BUF_SIZE);

	/* Fill in the URB for shipping it out. */
	odev->mux_bulk_tx_urb->transfer_flags = 0;
	usb_fill_bulk_urb(	odev->mux_bulk_tx_urb,
				odev->usb,
				usb_sndbulkpipe(odev->usb, odev->mux_ep_bulk_out),
				odev->mux_bulk_tx_buf,
				skb->len,
				write_bulk_callback,
				odev);

	/* Deal with the Zero Length packet problem, I hope */
	odev->mux_bulk_tx_urb->transfer_flags |= USB_ZERO_PACKET;

	/* Send the URB on its merry way. */
	if ((result = usb_submit_urb(odev->mux_bulk_tx_urb)))	{
		WARN("failed mux_bulk_tx_urb %d", result);
		odev->stats.tx_errors++;
		netif_start_queue(net);
	} else {
		odev->stats.tx_packets++;
		odev->stats.tx_bytes += skb->len;
		net->trans_start = jiffies;
	}
	dev_kfree_skb(skb);
	/* we're done */
	return result;
}

//=============================================================================
//	function:	hso_net_ioctl
//	purpose:	
//=============================================================================
static int hso_net_ioctl(struct net_device *net, struct ifreq *rq, int cmd)
{
	hso_priv *odev = net->priv;
	u32 usercmd = 0;
	char tmp[40];

	switch(cmd) {
		case SIOCDEVPRIVATE + 1: /* Chose this one because SIOCDEVPRIVATE used somewhere else in this code */
			/* return the number of sent bytes */
			D5("Transmitted: %lu", odev->stats.tx_bytes);
			rq->ifr_ifru.ifru_ivalue = odev->stats.tx_bytes;
			return 0;

		case SIOCETHTOOL:
			/* net specific */
			if (get_user(usercmd, (u32 *)rq->ifr_data))
				return -EFAULT;
		
			switch (usercmd) {
				case ETHTOOL_GDRVINFO: {
					/* get driver info */
					struct ethtool_drvinfo info = { ETHTOOL_GDRVINFO };
					strncpy(info.driver, driver_name, ETHTOOL_BUSINFO_LEN);
					strncpy(info.version, DRIVER_VERSION, ETHTOOL_BUSINFO_LEN);
					sprintf(tmp, "usb%d:%d", odev->usb->bus->busnum, odev->usb->devnum);
					strncpy(info.bus_info, tmp, ETHTOOL_BUSINFO_LEN);
					sprintf(tmp, "%s %x.%x", driver_name, ((odev->bcdCDC & 0xff00) >> 8), (odev->bcdCDC & 0x00ff) );
					strncpy(info.fw_version, tmp, ETHTOOL_BUSINFO_LEN);
					if (copy_to_user(rq->ifr_data, &info, sizeof(info)))
						return -EFAULT;

					return 0;
				}
				case ETHTOOL_GLINK: {
					/* get link status */
					struct ethtool_value edata = { ETHTOOL_GLINK };

					edata.data = netif_carrier_ok(net);
					if (copy_to_user(rq->ifr_data, &edata, sizeof(edata)))
						return -EFAULT;

					return 0;
				}
				default: {
					WARN("Got unsupported ioctl: %d", usercmd);
					return -EOPNOTSUPP;	/* the ethtool user space tool relies on this */
				}
			}
		default:
			return -ENOTTY; /* per ioctl man page */
	}
}

//=============================================================================
//	function:	hso_net_get_stats
//	purpose:	
//=============================================================================
static struct net_device_stats *hso_net_get_stats(struct net_device *net)
{
	return &((hso_priv *)net->priv)->stats;
}

//=============================================================================
//	function:	hso_net_tx_timeout
//	purpose:	called when a packet did not ack after watchdogtimeout
//=============================================================================
static void hso_net_tx_timeout(struct net_device *net)
{
	hso_priv *odev = net->priv;

	if (!odev) {
		return;
	}

	/* Tell syslog we are hosed. */
	WARN("%s: Tx timed out.", net->name);

	/* Tear the waiting frame off the list */
	odev->mux_bulk_tx_urb->transfer_flags |= USB_ASYNC_UNLINK;
	usb_unlink_urb( odev->mux_bulk_tx_urb );

	/* Update statistics */
	odev->stats.tx_errors++;
}

//=============================================================================
//	function:	hso_net_set_multicast
//	purpose:	setup the multicast filters
//=============================================================================
static void hso_net_set_multicast( struct net_device *net )
{
	hso_priv *odev = net->priv;
	int i = 0;
	__u8 *buf = NULL;

	// Tell the kernel to stop sending us frames while we get this all set up.
	netif_stop_queue(net);

	/* Note: do not reorder, GCC is clever about common statements. */
	if (net->flags & IFF_PROMISC) {
		/* Unconditionally log net taps. */
		D1( "%s: Promiscuous mode enabled", net->name);
		odev->mode_flags = MODE_FLAG_ALL_MULTICAST | MODE_FLAG_DIRECTED | MODE_FLAG_BROADCAST | MODE_FLAG_MULTICAST | MODE_FLAG_PROMISCUOUS;
	} else if (net->mc_count > odev->wNumberMCFilters) {
		/* Too many to filter perfectly -- accept all multicasts. */
		D1("%s: too many MC filters for hardware, using allmulti", net->name);
		odev->mode_flags = MODE_FLAG_ALL_MULTICAST | MODE_FLAG_DIRECTED | MODE_FLAG_BROADCAST | MODE_FLAG_MULTICAST;
	} else if (net->flags & IFF_ALLMULTI) {
		/* Filter in software */
		D1("%s: using allmulti", net->name);
		odev->mode_flags = MODE_FLAG_ALL_MULTICAST | MODE_FLAG_DIRECTED | MODE_FLAG_BROADCAST | MODE_FLAG_MULTICAST;
	} else {
		/* do multicast filtering in hardware */
		struct dev_mc_list *mclist;
		D1("%s: set multicast filters", net->name);
		odev->mode_flags = MODE_FLAG_ALL_MULTICAST | MODE_FLAG_DIRECTED | MODE_FLAG_BROADCAST | MODE_FLAG_MULTICAST;
		if (!(buf = kmalloc(6 * net->mc_count, GFP_ATOMIC))) {
			ERR("No memory to allocate?");
			goto exit;
		}
		for (i = 0, mclist = net->mc_list; mclist && i < net->mc_count; i++, mclist = mclist->next) {
				memcpy(&mclist->dmi_addr, &buf[i * 6], 6);
		}
		kfree(buf);
	}

exit:
	/* Tell the kernel to start giving frames to us again. */
	netif_wake_queue(net);
}

//=============================================================================
//	function:	read_bulk_callback
//	purpose:	Moving data from usb to kernel (in interrupt state)
//=============================================================================
static void read_bulk_callback( struct urb *urb )
{
	hso_priv *odev = NULL;
	struct net_device *net = NULL;
	int count = 0;
	int result = 0;

	if(!urb) {
		ERR("URB is NULL");
		return;
	}

	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_URB_KILLED:
			D1("killed %d", urb->status);
			return;
		case USB_ST_TIMEOUT:
			D1("timeout? %d", urb->status);
			return;	
		default:
			D1("rx status %d", urb->status);
	};

	odev = urb->context;
	count = urb->actual_length;

	/* Sanity check */
	if (!odev || !test_bit(HSO_NET_RUNNING, &odev->flags)) {
		D1("BULK IN callback but driver is not active!");
		return;
	}
	net = odev->net;

	if (!netif_device_present(net)) {
		/* Somebody killed our network interface... */
		return;
	}

	// STTH - NOTE: This will stall the reception of packets. FIX it if possible to get in this situation. 
	if ( test_and_set_bit(HSO_NET_RX_BUSY, &odev->flags)) {
		// Are we already trying to receive a frame???
		odev->stats.rx_errors++;
		ERR("odev Rx busy");
		return;
	}

	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_NORESPONSE:
			D1( "no repsonse in BULK IN" );
			clear_bit(HSO_NET_RX_BUSY, &odev->flags);
			break;
		default:
			D1( "%s: RX status %d", net->name, urb->status );
			goto goon;
	}

	// Check to make sure we got some data...
	if ( !count ) {
		// We got no data!!!
		D("no data");
		goto goon;
	}


	/* Handle the IP stream, add header and push it onto network stack if the packet is complete. */
	packetizeRx(odev, urb->transfer_buffer, count, (urb->transfer_buffer_length > count)?1:0 );

goon:
	/* We are done with this URB, resubmit it. */
	// Prep the USB to wait for another frame. Reuse same as received.
	usb_fill_bulk_urb(	urb,
				odev->usb,
				usb_rcvbulkpipe(odev->usb, odev->mux_ep_bulk_in),
				urb->transfer_buffer,
				MUX_BULK_RX_BUF_SIZE,
				read_bulk_callback,
				odev );

	urb->transfer_flags |= USB_QUEUE_BULK;

	// Give this to the USB subsystem so it can tell us
	// when more data arrives.
	if ( (result = usb_submit_urb(urb)) ) {
		WARN("%s failed submit modem_bulk_rx_urb %d", __FUNCTION__, result);
	}

	// We are no longer busy, show us the frames!!!
	clear_bit(HSO_NET_RX_BUSY, &odev->flags);
}

//=============================================================================
//	function:	packetizeRx
//	purpose:	make a real packet from the received USB buffer
//=============================================================================
static void packetizeRx(hso_priv *odev, unsigned char *ip_pkt, unsigned int count, unsigned char is_eop)
{
	unsigned short temp_bytes = 0, buffer_offset = 0, frame_len;
	unsigned char *tmp_rx_buf;
	struct  ethhdr *eth_head;
	
	while(count)
	{
		switch(odev->rx_parse_state)
		{
			case WAIT_IP: /* waiting for IP header. */        
				/* wanted bytes - size of ip header */
				temp_bytes = (count < odev->rx_buf_missing) ? count : odev->rx_buf_missing;
		
				memcpy(((unsigned char *)(&odev->rx_ip_hdr)) + odev->rx_buf_size, ip_pkt + buffer_offset, temp_bytes);
		
				odev->rx_buf_size       += temp_bytes;
				buffer_offset           += temp_bytes;
				odev->rx_buf_missing    -= temp_bytes;
				count                   -= temp_bytes;
		
				if(!odev->rx_buf_missing) { /* header is complete allocate an sk_buffer and continue to WAIT_DATA */
					frame_len = ntohs(odev->rx_ip_hdr.tot_len);
			
					if( frame_len > DEFAULT_MRU ) {
						odev->rx_parse_state = WAIT_SYNC;
						//printk(KERN_ERR "WAIT_SYNC: packet > MRU\n");
									continue;
					}
					/* Allocate an sk_buff */
					if ( !(odev->skb_rx_buf = dev_alloc_skb(frame_len + 2 + odev->net->hard_header_len)) ) {
						/* We got no receive buffer. */
						D("could not allocate memory");
						odev->rx_parse_state = WAIT_SYNC;
						//printk(KERN_ERR "WAIT_SYNC: Failed to allocate packet\n");
						return;
					}
					odev->skb_rx_buf->dev = odev->net;    /* Here's where it came from */
					
					/* Make some headroom: standard alignment + the ethernet header. */
					skb_reserve(odev->skb_rx_buf, 2 + odev->net->hard_header_len ); 
			
					/* Copy what we got so far. */
					tmp_rx_buf = skb_put(odev->skb_rx_buf, sizeof(struct iphdr));    /* make room for iphdr after tail. */
					memcpy(tmp_rx_buf, (char *)&(odev->rx_ip_hdr), sizeof(struct iphdr));
			
					odev->rx_buf_size = odev->net->hard_header_len + sizeof(struct iphdr); /* ETH_HLEN */
					odev->rx_buf_missing = frame_len - sizeof(struct iphdr); /* Filip actually use .tot_len */
					odev->rx_parse_state = WAIT_DATA;
				}
				break;

			case WAIT_DATA:
				temp_bytes = ( count < odev->rx_buf_missing ) ? count : odev->rx_buf_missing;

				/* copy the rest of the bytes that are left in the buffer into the waiting sk_buf */
				tmp_rx_buf = skb_put(odev->skb_rx_buf, temp_bytes);    /* make room for temp_bytes after tail. */
				memcpy(tmp_rx_buf, ip_pkt + buffer_offset, temp_bytes);

				odev->rx_buf_missing    -= temp_bytes;
				count                   -= temp_bytes;
				buffer_offset           += temp_bytes;
				odev->rx_buf_size       += temp_bytes;
				if( !odev->rx_buf_missing ) { // Packet is complete. Inject into stack.
					/* Add fake ethernet header. */
					eth_head = (struct ethhdr *)skb_push(odev->skb_rx_buf, odev->net->hard_header_len); /* decrease headroom */
					memcpy(eth_head, &odev->dummy_eth_head, sizeof(struct ethhdr));

					// Not sure here either
					odev->skb_rx_buf->protocol  = eth_type_trans(odev->skb_rx_buf, odev->net);
					odev->skb_rx_buf->ip_summed = CHECKSUM_UNNECESSARY;         /* don't check it */

					//printk( KERN_ERR "RX PKT %d bytes\n", odev->rx_buf_size );
					// Ship it off to the kernel
					netif_rx(odev->skb_rx_buf);
					odev->skb_rx_buf = NULL;            /* No longer our buffer. */

					// update out statistics
					odev->stats.rx_packets ++;
					odev->stats.rx_bytes += odev->rx_buf_size;      /* Hmmm, wonder if we have recieved the IP len or the ETH len. */

					odev->rx_buf_size = 0;
					odev->rx_buf_missing = sizeof(struct iphdr);
					odev->rx_parse_state = WAIT_IP;
				}
				break;

			case WAIT_SYNC:
				D(" W_S");
				count = 0;
				break;
			default:
				D(" ");
				count --; 
				break;
		}
	}

	// Recovery mechanism for WAIT_SYNC state. Can't believe he didn't port this over.
	if( is_eop ) {
		if( odev->rx_parse_state == WAIT_SYNC ) {
			odev->rx_parse_state = WAIT_IP;
			odev->rx_buf_size = 0;
			odev->rx_buf_missing = sizeof(struct iphdr);
		} 
	}
}

//=============================================================================
//	function:	write_bulk_callback
//	purpose:	USB tells is xmit done, we should start the netqueue again
//=============================================================================
static void write_bulk_callback( struct urb *urb )
{
	hso_priv *odev = NULL;

	if(!urb) {
		ERR( "URB is NULL" );
		return;
	}

	odev = urb->context;
	// Sanity check
	if( !odev || !test_bit(HSO_NET_RUNNING, &odev->flags)) {
		// We are insane!!!
		ERR( "write_bulk_callback: device not running" );
		return;
	}

	// Do we still have a valid kernel network device?
	if ( !netif_device_present(odev->net) ) {
		// Someone killed our network interface.
		ERR( "write_bulk_callback: net device not present" );
		return;
	}

	// Hmm...  What on Earth could have happened???
	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_URB_KILLED:
			D1("killed %d", urb->status);
			return;
		case USB_ST_TIMEOUT:
			D1("timeout? %d", urb->status);
			return;	
		default:
			D1("tx status %d", urb->status);
	};

	// Update the network interface and tell it we are
	// ready for another frame
	odev->net->trans_start = jiffies;
	netif_wake_queue( odev->net );
}

//=============================================================================
//	function:	hso_handle_arp
//	purpose:	returns 1 - if the packet is an ARP packet and the original packet is dropped.
//				0 - packet is not an ARP packet.
//=============================================================================
static int hso_handle_arp( struct sk_buff *skb, struct net_device *net ) 
{
	unsigned char  t_ip[4], s_ip[4];
	unsigned char *buf2, sha[6], tha[6];
	struct arphdr *arp_req;

	/* Not an ARP packet. */
	if ((skb->protocol) != ntohs(ETH_P_ARP)) {
		return 0;
	}

	D1(" protocol = %04x", skb->protocol);

	/* Else what.. */
	arp_req = skb->nh.arph;

	DUMP1(skb->data, skb->len);

	/* Wrong hardware type - only handling ethernet. */
	if(ntohs(arp_req->ar_hrd) != ARPHRD_ETHER)
		return 0;

	/* Not a request. lets ignore it. */
	if(ntohs(arp_req->ar_op) != ARPOP_REQUEST)
		goto drop;

	/* extract data from request */
	buf2 = (unsigned char *)skb->nh.arph;
	buf2 += sizeof(struct arphdr);

	memcpy (sha, buf2, ETH_ALEN);
	buf2 += ETH_ALEN;

	memcpy (s_ip, buf2, 4);
	buf2 += 4;

	memcpy (tha, buf2, ETH_ALEN);
	buf2 += ETH_ALEN;

	memcpy (t_ip, buf2, 4);

	arp_send_response(	net,
				net->dev_addr,			//unsigned char *dst_hw
				(unsigned char *)dummy_mac,	//unsigned char *src_hw
				*(u32*)t_ip,			// u32 s_pa
				*(u32*)s_ip,			// u32 t_pa
				(unsigned char *)dummy_mac,	// unsigned char *s_ha
				sha);				// unsigned char *t_ha

drop:
	dev_kfree_skb(skb);
	return 1;
}

//=============================================================================
//	function:	arp_send_response
//	purpose:	This functions creates a new packet to use for sending an ARP reply,
//			emulating a received packet it is "sent" through netif_rx.
//=============================================================================
static void arp_send_response( struct net_device *dev, unsigned char *dst_hw, unsigned char *src_hw, u32 s_pa,u32 t_pa, unsigned char *s_ha, unsigned char *t_ha) 
{
	struct sk_buff *skb = NULL;
	int count = sizeof(struct arphdr)+ 2*(dev->addr_len+4) + dev->hard_header_len;
	char buf[60];
	char *p = NULL;
	struct ethhdr *eth_head = NULL;
	struct arphdr *arp_req = NULL;

	D1("-----------------------------------");

	if ( !netif_device_present(dev) ) {
		// Somebody killed our network interface...
		return;
	}

	skb = alloc_skb(count + 2, GFP_ATOMIC);
	if (skb == NULL) {
		return;
	}
	skb_reserve(skb, 2);

	// Here's where it came from
	skb->dev = dev;

	p = buf;
	eth_head = (struct ethhdr*)buf;

	memcpy(eth_head->h_dest, dst_hw, ETH_ALEN);     /* driver MAC */
	memcpy(eth_head->h_source, src_hw , ETH_ALEN);  /* from dummy device */
	eth_head->h_proto = htons(ETH_P_ARP);

	arp_req = (struct arphdr *)(eth_head+1);
	arp_req->ar_hrd = htons(ARPHRD_ETHER);
	arp_req->ar_pro = htons(ETH_P_IP);
	arp_req->ar_hln = 6;
	arp_req->ar_pln = 4;
	arp_req->ar_op =  htons(ARPOP_REPLY);

	p = (char *)(arp_req+1);

	// Now we copy it over
	memcpy(p , s_ha, ETH_ALEN);
	p += ETH_ALEN;

	memcpy(p , &s_pa, 4);
	p += 4;

	memcpy(p , t_ha, ETH_ALEN);
	p += ETH_ALEN;

	memcpy(p , &t_pa, 4);

	// Not sure
	memcpy(skb_put(skb, count), buf, count);

	// Not sure here either
	skb->protocol = eth_type_trans(skb, dev);
	skb->ip_summed = CHECKSUM_UNNECESSARY; /* don't check it */

	DUMP1(skb->data, skb->len);

	// Ship it off to the kernel
	netif_rx(skb);
}


//*****************************************************************************
//*****************************************************************************
// Serial driver functions
//*****************************************************************************
//*****************************************************************************

//=============================================================================
//	function:	hso_serial_open
//	purpose:	open the requested serial port
//=============================================================================
static int hso_serial_open(struct tty_struct *tty, struct file * filp)
{
	hso_serial_t *serial = NULL;
	int minor = MINOR(tty->device);

	D1("Opening %d", minor);

	serial = hso_serial_table[minor]; /* initially all NULL */
	if(serial == NULL || serial->magic != HSO_SERIAL_MAGIC) {
		ERR("Failed to open port");
		return -ENODEV;
	}
	if(serial->open_count > 0) {
		++serial->open_count;
		D1("Port was already open");
		return 0;
	}

	spin_lock( &serial->serial_lock );

	tty->driver_data = serial;
	serial->tty = tty;

	++serial->open_count;
	if (serial->open_count == 1) {
		tty->low_latency = 1;
		serial->flags = 0;
		serial->read_scheduled = 0;

		/* Force default termio settings */
		_hso_serial_set_termios (tty, NULL) ;

		/* Rx fifo init. */
		serial->rx_fifo.p_h = serial->rx_fifo.p_t = serial->rx_fifo.p_b = serial->rx_buffer;
		serial->rx_fifo.cnt = 0;
		serial->rx_fifo.size = sizeof(serial->rx_buffer);

		serial->rx_flip_task.list.next = NULL;
		serial->rx_flip_task.list.prev = NULL;
		serial->rx_flip_task.sync = 0;
		serial->rx_flip_task.data = (void *)tty;
		serial->rx_flip_task.routine = hso_serial_rx_flip;
	}

	spin_unlock( &serial->serial_lock );

	/* If it is the diagnostics port fill in and submit a bulk urb. (already allocated in hso_serial_start) */
	if( serial->type != MUX_INTERFACE) {
		D("Submitting read urb for serial port %d", serial->minor);
		int res = 0;
		usb_fill_bulk_urb(	serial->rx_urb, 
					serial->odev->usb,
					usb_rcvbulkpipe(serial->odev->usb, serial->ep_bulk_in),
					serial->rx_data,
					serial->rx_data_length,
					hso_std_serial_read_bulk_callback,
					serial);
		serial->rx_urb->transfer_flags |= USB_QUEUE_BULK;
		if((res = usb_submit_urb(serial->rx_urb))) {
			D("Failed to submit urb - res %d", res);
		}
	}

	return 0;
}

//=============================================================================
//	function:	hso_serial_close
//	purpose:	close the requested serial port
//=============================================================================
static void hso_serial_close (struct tty_struct *tty, struct file * filp)
{
	hso_serial_t *serial = NULL;
	D1(" ");

	if(tty == NULL || tty->driver_data == NULL) {
		return;
	}

	serial = tty->driver_data;
	spin_lock_bh( &serial->serial_lock );
	if (tty->driver_data) {
		__hso_serial_close(serial);
	}
	spin_unlock_bh( &serial->serial_lock );
}

//=============================================================================
//	function:	hso_serial_write
//	purpose:	close the requested serial port
//=============================================================================
static int hso_serial_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count)
{
	hso_serial_t *serial = get_serial(tty);
	int space = 0, tx_bytes = 0;

	if(serial == NULL) {
		ERR("%s(%d) tty or tty->driver_data is NULL", __FUNCTION__ , __LINE__);
		return -ENODEV;
	}

	spin_lock_bh( &serial->serial_lock );
	space = serial->tx_data_length - serial->tx_buffer_count;
	tx_bytes = (count < space )? count : space;

	if( !tx_bytes ) {
		spin_unlock_bh( &serial->serial_lock );
		return 0;
	}

	D2(" hso_port = %d count %d", serial->minor , tx_bytes);

	if (from_user) {
		if (copy_from_user(serial->tx_buffer + serial->tx_buffer_count,  buf, tx_bytes)) {
			clear_bit(HSO_PRIV_FLAG_TX_SENT, &serial->flags);
			spin_unlock_bh( &serial->serial_lock );
			D2("Could not copy data");
			return -EFAULT;
		}
		serial->tx_buffer_count+= tx_bytes;
	} else {
		memcpy( serial->tx_buffer + serial->tx_buffer_count,  buf, tx_bytes );
		serial->tx_buffer_count+= tx_bytes;
	}
	spin_unlock_bh( &serial->serial_lock );
	DUMP1(buf, count);

	/* send it */
	hso_kick_transmit( serial );
	/* done */
	return tx_bytes;
}

//=============================================================================
//	function:	hso_serial_write_room
//	purpose:	how much room is there for writing
//=============================================================================
static int hso_serial_write_room (struct tty_struct *tty)
{
	hso_serial_t *serial = get_serial(tty);
	int room = 0;

	/* sanity check */
	if (serial == NULL) {
		return 0;
	}

	spin_lock_bh( &serial->serial_lock );
	room = serial->tx_data_length - serial->tx_buffer_count;
	spin_unlock_bh( &serial->serial_lock );

	/* return free room */
	return room;
}

//=============================================================================
//	function:	hso_serial_set_termios
//	purpose:	setup the term
//=============================================================================
static void hso_serial_set_termios (struct tty_struct *tty, struct termios * old)
{
	hso_serial_t *serial = get_serial(tty);

	if ((!tty) || (!tty->termios) || (!serial)) {
		D1("no tty structures");
		return;
	}

	D1("port %d", serial->minor);

	spin_lock_bh( &serial->serial_lock );
	if (serial->open_count) {
		_hso_serial_set_termios(tty, old);
	}
	spin_unlock_bh( &serial->serial_lock );

	return;
}

//=============================================================================
//	function:	hso_serial_chars_in_buffer
//	purpose:	how many characters in the buffer
//=============================================================================
static int hso_serial_chars_in_buffer (struct tty_struct *tty)
{
	hso_serial_t *serial = get_serial(tty);
	int chars = 0;

	/* sanity check */
	if(serial == NULL) {
		return 0;
	}

	spin_lock_bh( &serial->serial_lock );
	chars = serial->tx_buffer_count;
	spin_unlock_bh( &serial->serial_lock );

	return chars;
}

//=============================================================================
//	function:	hso_serial_ioctl
//	purpose:	ioctl not supported
//=============================================================================
static int hso_serial_ioctl (struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg)
{
	D4(" ");
	return -ENOIOCTLCMD;
}

//=============================================================================
//	function:	hso_serial_throttle
//	purpose:	throttle not supported
//=============================================================================
static void hso_serial_throttle (struct tty_struct * tty)
{
	D1(" ");
}

//=============================================================================
//	function:	hso_serial_unthrottle
//	purpose:	throttle not supported
//=============================================================================
static void hso_serial_unthrottle (struct tty_struct * tty)
{
	D1(" ");
}

//=============================================================================
//	function:	hso_serial_break
//	purpose:	break not supported
//=============================================================================
static void hso_serial_break (struct tty_struct *tty, int break_state)
{
	D1(" ");
}

//=============================================================================
//	function:	hso_serial_read_proc
//	purpose:	read_proc not supported
//=============================================================================
static int hso_serial_read_proc (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	return 0;
}

//=============================================================================
//	function:	hso_serial_hangup
//	purpose:	hangup not supported
//=============================================================================
static void hso_serial_hangup (struct tty_struct *tty)
{
    D1("hang up");
}

//=============================================================================
//	function:	hso_kick_transmit
//	purpose:	starts a transmit
//=============================================================================
static void hso_kick_transmit( hso_serial_t *serial )
{
	u8	*temp = 0;
	unsigned long flags;

	if(!serial || serial->open_count <= 0) {
		D("No serial");
		return;
	}

	spin_lock_irqsave( &serial->serial_lock, flags );
	if( !serial->tx_buffer_count ) {
		D5("Buffer empty");
		goto out;
	}

	if( serial->tx_urb->status == -EINPROGRESS ) {
		D5("TX in progress");
		goto out;
	}
	if( serial->type == MUX_INTERFACE && serial->rx_urb->status == -EINPROGRESS ) {
		D5("RX in progress");
		goto out;
	}

	//Switch pointers around to avoid memcpy
	temp = serial->tx_buffer;
	serial->tx_buffer = serial->tx_data;
	serial->tx_data = temp;
	serial->tx_data_count = serial->tx_buffer_count;
	serial->tx_buffer_count = 0;

out:	
	spin_unlock_irqrestore( &serial->serial_lock, flags );
#if ( LINUX_VERSION_CODE > KERNEL_VERSION(2,4,33) )
	if(serial->tty) {
		tty_wakeup( serial->tty );
	} else {
		D("Could not wake tty");
	}
#endif
	//If temp is set, it means we switched buffers
	if( temp && serial->write_data ) {
		D5("Going to write");
		serial->write_data( serial );
	}
	D5("Kick transmit ready");
}

//=============================================================================
//	function:	hso_kick_receive
//	purpose:	starts a receive for the muxed port
//=============================================================================
static void hso_kick_receive( hso_serial_t *serial )
{
	if(!serial) {
		D("No serial");
		return;
	}

	if( serial->tx_urb->status == -EINPROGRESS || serial->rx_urb->status == -EINPROGRESS) {
		D5("Ctrl urb in progress");
		return;
	}

	if( serial->read_scheduled) {
		D5("Going to read");
		serial->read_scheduled = 0;
		hso_mux_serial_read(serial->odev, serial);
	}
	D5("Kick receive ready");
}

//=============================================================================
//	function:	intr_callback
//	purpose:	used for muxed serial port callback (muxed serial read)
//=============================================================================
static void intr_callback( struct urb *urb )
{
	hso_priv *odev = urb->context;
	struct net_device *net = NULL;
	unsigned char *port_req = NULL;
	int i = 0;

	if ( !odev ) {
		return;
	}

	D4("------------- got intr callback %d -------------------------------", urb->status);

	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_TIMEOUT:
		case USB_ST_URB_KILLED:
		default:
			D1("intr status %d", urb->status);
			return;
	};

	net = odev->net;
	port_req = urb->transfer_buffer;

	D4(" port_req = 0x%.2X", *port_req );

	for(i = 0; i < HSO_SERIAL_TTY_CTRL_DEVS ; i++) {
		if( (*port_req & (1<<i)) && hso_serial_table[i] != NULL ) {
			D1("Pending read interrupt on port %d", i);
			if(!test_and_set_bit(HSO_PRIV_FLAG_RX_SENT, &hso_serial_table[i]->flags)) {
				/* Setup and send send a ctrl req read on port i */
				hso_serial_table[i]->read_scheduled = 1;
				hso_kick_receive(hso_serial_table[i]);
			} else {
				D1("Already pending a read on port %d", i);
			}
		}
	}
}

//=============================================================================
//	function:	hso_mux_serial_read
//	purpose:	called by intr_callback when read occurs
//=============================================================================
static int hso_mux_serial_read(hso_priv *odev, hso_serial_t* serial)
{
	int rc = 0;

	memset(serial->rx_data, 0, CTRL_URB_RX_SIZE); /* rx_data_length */

	rc = mux_device_request(serial,
				GET_ENCAPSULATED_RESPONSE,
				serial->minor,
				serial->rx_urb,
				&serial->ctrl_req_rx,
				serial->rx_data,
				serial->rx_data_length);
	return rc;
}

//=============================================================================
//	function:	hso_mux_serial_write_data
//	purpose:	called for writing to muxed serial port
//=============================================================================
static int hso_mux_serial_write_data(hso_serial_t *serial)
{
    return mux_device_request(	serial,
				SEND_ENCAPSULATED_COMMAND,
				serial->minor,
				serial->tx_urb,
				&serial->ctrl_req_tx,
				serial->tx_data,
				serial->tx_data_count);
}

//=============================================================================
//	function:	hso_std_serial_write_data
//	purpose:	called for writing diag or CS serial port
//=============================================================================
static int hso_std_serial_write_data(hso_serial_t *serial)
{
	int count = serial->tx_data_count;
	int res = 0;

	usb_fill_bulk_urb(	serial->tx_urb,
				serial->odev->usb,
				usb_sndbulkpipe(serial->odev->usb, serial->ep_bulk_out),
				serial->tx_data,
				serial->tx_data_count,
				hso_std_serial_write_bulk_callback,
				serial);
	serial->tx_urb->transfer_flags |= USB_QUEUE_BULK;

	if((res = usb_submit_urb(serial->tx_urb)))
	{
		D("Failed to submit urb - res %d status %d ", res, serial->tx_urb->status);
		clear_bit(HSO_PRIV_FLAG_TX_SENT, &serial->flags);
		return res;
	}

	return count;
}

//=============================================================================
//	function:	mux_device_request
//	purpose:	make a request (for reading and writing data to muxed serial port)
//=============================================================================
static int mux_device_request(hso_serial_t *serial, u8 type, u16 port, struct urb* ctrl_urb, struct usb_ctrlrequest *ctrl_req, u8* ctrl_urb_data, u32 size)
{
	int res = 0;
	int pipe = -1;
	int counter = 0;

	// Sanity check
	if (!serial || !ctrl_urb || !ctrl_req) {
		ERR("Wrong arguments ");
		return -EINVAL;
	}

	ctrl_req->wValue = 0;
	ctrl_req->wIndex = port; // port to send to
	ctrl_req->wLength = size; // Is this supposed to be this???

	if (type == GET_ENCAPSULATED_RESPONSE) {
		// Reading command
		ctrl_req->bRequestType	= USB_DIR_IN | USB_TYPE_OPTION_VENDOR | USB_RECIP_INTERFACE;
		ctrl_req->bRequest		= GET_ENCAPSULATED_RESPONSE;
		pipe = usb_rcvctrlpipe(serial->odev->usb, 0);
	} else {
		// Writing command
		ctrl_req->bRequestType	= USB_DIR_OUT | USB_TYPE_OPTION_VENDOR | USB_RECIP_INTERFACE;
		ctrl_req->bRequest		= SEND_ENCAPSULATED_COMMAND;
		pipe = usb_sndctrlpipe(serial->odev->usb, 0);
	}

	D2("%s command (%02x) len: %d, port: %d", type == GET_ENCAPSULATED_RESPONSE ? "Read":"Write", ctrl_req->bRequestType, ctrl_req->wLength, port);

	if(!in_interrupt()) {
		while(serial->rx_urb->status == -EINPROGRESS || serial->tx_urb->status == -EINPROGRESS) {
			if(counter++ > 20) {
				ERR("Timeout exceeded while trying to write to serial port /dev/ttyHS%d", serial->minor);
				return -EINVAL;
			}
			D2("Waiting for previous ctrl urb to complete...");
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,33) )
			udelay(100);
#else
			msleep(100);
#endif
		}
	}

	// Load ctrl urb
	ctrl_urb->transfer_flags = 0;
	usb_fill_control_urb(ctrl_urb, serial->odev->usb, pipe, (u8*) ctrl_req, ctrl_urb_data, size, ctrl_callback, serial);

	// Send it on merry way
	if ( (res = usb_submit_urb(ctrl_urb)) ) {
		ERR("%s failed submit ctrl_urb %d type %d status %d", __FUNCTION__, res, type, ctrl_urb->status);
		return res;
	}

	return size;
}

//=============================================================================
//	function:	ctrl_callback
//	purpose:	callback after read or write on muxed serial port
//=============================================================================
static void ctrl_callback( struct urb *urb )
{
	hso_serial_t *serial = urb->context;
	struct usb_ctrlrequest *req = NULL;

	if ( !serial ) {
		return;
	}

	req = (struct usb_ctrlrequest*) (urb->setup_packet);

	D4("-------------------- got ctrl callback %02x ---------------------", urb->status);

        /* This is old values */
	/*if (req) {
		D4("bRequestType: 0x%02x", req->bRequestType);
		D4("bRequest:	 0x%02x", req->bRequest);
		D4("wValue:       0x%04x", req->wValue);
		D4("wIndex:       0x%04x", req->wIndex);
		D4("wLength:      0x%04x (%d)", req->wLength, req->wLength);
	}*/

	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_URB_KILLED:
			D1("--");
			return;
		default:
			D1("intr status %d", urb->status);
	};

	if (req->bRequestType == (USB_DIR_IN | USB_TYPE_OPTION_VENDOR | USB_RECIP_INTERFACE)) {
		D4("Ctrl callback in response to read command");
		if(serial->open_count > 0 && urb->status == USB_ST_NOERROR) {
			put_rxbuf_data(urb, serial);
		}

		/* Re issue a read as long as we receive data. */
		if(urb->actual_length != 0) {
			serial->read_scheduled = 1;
			hso_kick_receive( serial );
		} else {
			clear_bit(HSO_PRIV_FLAG_RX_SENT, &serial->flags);
			// if no read is there, but write is there, better do a transmit again
			hso_kick_transmit(serial);
		}
	} else { /* write request. */
		D4("Ctrl callback in response to write command");
		hso_kick_transmit( serial );
		hso_kick_receive( serial );
		// queue_task(&serial->tqueue, &tq_immediate);
		// mark_bh(IMMEDIATE_BH);
	}
}

//=============================================================================
//	function:	put_rxbuf_data
//	purpose:	handle RX data for serial port
//=============================================================================
static void put_rxbuf_data(struct urb *urb, hso_serial_t *serial)
{
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,34) )
	unsigned char *data = NULL;
	fifo_t *rx_fifo = NULL;
	int n = 0;
	int endlen = 0;

	if(urb == NULL || serial == NULL) {
		D1("serial = NULL");
		return;
	}

	data  = urb->transfer_buffer;
	rx_fifo = &serial->rx_fifo;
	n = urb->actual_length;

	if(rx_fifo == NULL) {
		D1("fifo = NULL");
		return;
	}

	/* check room ... */
	if (urb->actual_length > (((rx_fifo->size)-1) - rx_fifo->cnt)) {
		ERR("%s - no room in rx buffer, discarding data", __FILE__);
		return;
	}

	endlen = &rx_fifo->p_b[(rx_fifo->size)-1] - rx_fifo->p_h + 1 ;

	if ( n < endlen ) {
		memcpy(rx_fifo->p_h, data, n );
		rx_fifo->p_h +=n;
	} else {
		memcpy(rx_fifo->p_h , data, endlen);
		memcpy(rx_fifo->p_b, data + endlen, n - endlen);
		rx_fifo->p_h = rx_fifo->p_b + (n - endlen);
	}
	rx_fifo->cnt += n;
	D1("%s - port %d - rx URB, len %d, bytes in buffer %d", __FUNCTION__, serial->minor, urb->actual_length, rx_fifo->cnt);

	if ( rx_fifo->cnt >= HSO_THRESHOLD_THROTTLE	 && !test_and_set_bit(HSO_PRIV_FLAG_THROTTLED, &serial->flags) ) {
		D1("Throttle: rx_fifo->cnt[%d] >= HSO_THRESHOLD_THROTTLE[%d]", rx_fifo->cnt, HSO_THRESHOLD_THROTTLE);
	}
	schedule_task(&serial->rx_flip_task);
#else	/* ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,34) ) */


        struct tty_struct *tty = serial->tty;

	/* Sanity check */
	if (urb == NULL || serial == NULL) {
		D1("serial = NULL");
		return;
	}

       /* Push data to tty */
        if (tty && urb->actual_length) {
		D("data to push to tty");

		int i = 0;
		unsigned char *data = urb->transfer_buffer;
		DUMP1(data, urb->actual_length);

		for (i = 0; i < urb->actual_length ; ++i) {
			if (tty->flip.count >= TTY_FLIPBUF_SIZE)
				tty_flip_buffer_push(tty);
			tty_insert_flip_char(tty, data[i], 0);
		}
		tty_flip_buffer_push(tty);
        }


#endif	/* ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,34) ) */
}

//=============================================================================
//	function:	hso_std_serial_read_bulk_callback
//	purpose:	read callback for Diag and CS port
//=============================================================================
static void hso_std_serial_read_bulk_callback( struct urb *urb )
{
	hso_serial_t *serial = urb->context;
	int res = 0;

	if ( !serial ) {
		return;
	}

	D4("-------------------- got serial_read_bulk callback %02x ---------------------", urb->status);

	switch ( urb->status ) {
		case USB_ST_NOERROR:
			break;
		case USB_ST_URB_KILLED:
			D1("--");
			break;
		case -ETIMEDOUT:
			return;
		default:
			D1("urb status %d", urb->status);
			break;
	};

	D1("actual length = %d", urb->actual_length);
	DUMP1(urb->transfer_buffer, urb->actual_length);

	if(serial->open_count == 0) { //No one is listening, dont resubmit it..
		return;
	}

	if(urb->status == USB_ST_NOERROR) { /* Valid data */
		put_rxbuf_data(urb, serial);
	} else if (urb->status == -ENOENT || urb->status == -ECONNRESET) { /* Unlinked - check for throttled port. */
		D2(" port %d, successfully unlinked urb", serial->minor);        
	} else {
		D2(" port %d, status = %d for read urb", serial->minor, urb->status);
		return ;
	}

	/* We are done with this URB, resubmit it. */
	// Prep the USB to wait for another frame
	usb_fill_bulk_urb(	serial->rx_urb,
				serial->odev->usb,
				usb_rcvbulkpipe(serial->odev->usb, serial->ep_bulk_in),
				serial->rx_data,
				serial->rx_data_length,
				hso_std_serial_read_bulk_callback,
				serial );

	serial->rx_urb->transfer_flags |= USB_QUEUE_BULK;

	// Give this to the USB subsystem so it can tell us when more data arrives.
	if ( (res = usb_submit_urb(serial->rx_urb)) ) {
		ERR("%s failed submit serial rx_urb %d status %d", __FUNCTION__, res, serial->rx_urb->status);
	}
}

//=============================================================================
//	function:	hso_std_serial_write_bulk_callback
//	purpose:	write callback for Diag and CS port
//=============================================================================
static void hso_std_serial_write_bulk_callback( struct urb *urb ) 
{
	hso_serial_t *serial = (hso_serial_t *)urb->context;

	if(!serial)
		return;

	clear_bit(HSO_PRIV_FLAG_TX_SENT, &serial->flags);

	if (urb->status) {
		D1("%s - nonzero write bulk status received: %d", __FUNCTION__, urb->status);
		return;
	}

	hso_kick_transmit( serial );
	// queue_task(&serial->tqueue, &tq_immediate);
	// mark_bh(IMMEDIATE_BH);
	return;
}

//=============================================================================
//	function:	_hso_serial_set_termios
//	purpose:	
//=============================================================================
static void _hso_serial_set_termios (struct tty_struct *tty, struct termios * old)
{
	hso_serial_t *serial = get_serial(tty);

	if ((!tty) || (!tty->termios) || (!serial)) {
		ERR("no tty structures");
		return;
	}

	D4("port %d", serial->minor);

	/*
	 * The default requirements for this device are:
	 */
	serial->tty->termios->c_iflag &=~(IGNBRK	/* disable ignore break */
					| BRKINT	/* disable break causes interrupt */
					| PARMRK	/* disable mark parity errors */
					| ISTRIP	/* disable clear high bit of input characters */
					| INLCR		/* disable translate NL to CR */
					| IGNCR		/* disable ignore CR */
					| ICRNL		/* disable translate CR to NL */
					| IXON);	/* disable enable XON/XOFF flow control */

	serial->tty->termios->c_oflag &= ~OPOST;	/* disable postprocess output characters */

	serial->tty->termios->c_lflag &= ~(ECHO		/* disable echo input characters */
					| ECHONL	/* disable echo new line */
					| ICANON	/* disable erase, kill, werase, and rprnt special characters */
					| ISIG		/* disable interrupt, quit, and suspend special characters */
					| IEXTEN);	/* disable non-POSIX special characters */

	serial->tty->termios->c_cflag &= ~(CSIZE	/* no size */
					| PARENB	/* disable parity bit */
					| CBAUD);	/* clear current baud rate */

	serial->tty->termios->c_cflag |= (CS8		/* character size 8 bits */
					| B115200);	/* baud rate 115200 */

	/*
	 * Force low_latency on; otherwise the pushes are scheduled;
	 * this is bad as it opens up the possibility of dropping bytes
	 * on the floor.  We don't want to drop bytes on the floor. :)
	 */
	serial->tty->low_latency = 1;

	/* Notify the tty driver that the termios have changed. */
	serial->tty->ldisc.set_termios(serial->tty, NULL);
	return;
}


//*****************************************************************************
//*****************************************************************************
// Procfs functions
//*****************************************************************************
//*****************************************************************************

//=============================================================================
//	function:	hso_read_info
//	purpose:	called when reading the procfile "hso"
//=============================================================================
int hso_read_info(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	
	D1("count: %d", count);	
	len += snprintf(buf+len, count-len, "debug: %02x", debug);
	return len;
}


//*****************************************************************************
//*****************************************************************************
// Base driver functions
//*****************************************************************************
//*****************************************************************************

//=============================================================================
//	function:	hso_init
//	purpose:	called when loading the module, initializes serial
//			interface, registers usb driver and proc devices
//=============================================================================
int __init hso_init(void) {
	int rc = 0;
	D1("%s", version );
	if ((rc = hso_serial_init())) {
		ERR("Could not register hso driver? error: %d", rc);
		return rc;
	}

	create_proc_read_entry("driver/hso", 0, NULL, hso_read_info, NULL);

	return usb_register( &hso_driver );
}

//=============================================================================
//	function:	hso_exit
//	purpose:	called when unloading the module, cleans up
//=============================================================================
void __exit hso_exit(void) {
	D1(" CC ");
	remove_proc_entry("driver/hso", NULL);

	usb_deregister( &hso_driver );
	hso_serial_exit();
}

//=============================================================================
//	function:	hso_probe
//	purpose:	called once for each interface upon device insertion */
//=============================================================================
static void* hso_probe(struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id)
{
	struct net_device *net = NULL;
	hso_priv *odev = NULL;
	int rc = 0;
	int ifnr = 0;
	int i = 0;
	int res = 0;

	// First we should check the active configuration to see if
	// any other driver has claimed any of the interfaces.
	// Or if this is the second time we are probed.
	if ( global_odev && check_for_claimed_interfaces( usb->actconfig, global_odev ) ) {
		WARN("Someone (we) has claimed our interface, bail out");
		return NULL;
	}

	// We might be finding a device we can use.
	// We all go ahead and allocate our storage space.
	// We need to because we have to start filling in the data that
	// we are going to need later.
	if (hso_alloc_memory(&odev)) {
		ERR("Could not allocate memory?");
		goto exit;
	}
	global_odev = odev;

	// Init interfaces for sanity check
	odev->interfaces[MUX_INTERFACE] = NULL;
	odev->interfaces[QXDM_INTERFACE] = NULL;
	odev->interfaces[CS_INTERFACE] = NULL;

	// Let's see if we can find a configuration we can use.
	if ((rc = find_valid_configuration( usb, odev ))) {
		// Nope we couldn't find one we liked.
		// This device was not meant for us to control.
		D1("Could not find valid configuration");
		goto exit;
	}

	// Now that we FOUND a configuration. let's try to make the
	// device go into it.
	if ( usb_set_configuration( usb, odev->bConfigurationValue ) ) {
		ERR("usb_set_configuration() failed");
		goto exit;
	}

	// Now set the communication interface up as required.
	if (usb_set_interface(usb, odev->mux_bInterfaceNumber, odev->mux_altsetting)) {
		ERR("usb_set_interface() failed");
		goto exit;
	}

	// Now we need to get a kernel Ethernet interface.
	net = init_etherdev( NULL, 0 );
	if ( !net ) {
		ERR( "Unable to initialize ethernet device" );
		goto exit;
	}
	odev->skb_rx_buf = NULL;
	odev->rx_parse_state = WAIT_IP;
	// Now that we have an ethernet device, let's set it up
	// (And I don't mean "set [it] up the bomb".)
	odev->wMaxSegmentSize = DEFAULT_MTU;

	net->priv = odev;
	SET_MODULE_OWNER(net);
	net->open = hso_net_open;
	net->stop = hso_net_close;
	net->watchdog_timeo = HSO_NET_TX_TIMEOUT;
	net->tx_timeout = hso_net_tx_timeout;	// TX timeout function
	net->do_ioctl = hso_net_ioctl;
	net->hard_start_xmit = hso_net_start_xmit;
	net->set_multicast_list = hso_net_set_multicast;
	net->get_stats = hso_net_get_stats;
	net->mtu = odev->wMaxSegmentSize - 14;

	// We'll keep track of this information for later...
	odev->usb = usb;
	odev->net = net;

	// and don't forget the MAC address.
	set_ethernet_addr( odev );

	// Send a message to syslog about what we are handling
	log_device_info( odev );

	/* We need to manually claim the data interface, while the comm interface gets claimed in the return */
	usb_driver_claim_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[MUX_INTERFACE]), odev );
	usb_driver_claim_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[QXDM_INTERFACE]), odev );
	usb_driver_claim_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[CS_INTERFACE]), odev );

	D1("enabled ports: %0x", odev->option_info.enabled_ports);
	for(i = 1; i < 0x100; i = i << 1, ifnr++) {
		if(odev->option_info.enabled_ports & i) {
			last_serial_minor_used++;
			hso_serial_start(odev, last_serial_minor_used, MUX_INTERFACE, ifnr);
		}
	}

	/* Diagnostics port */
	last_serial_minor_used++;
	hso_serial_start(odev, last_serial_minor_used, QXDM_INTERFACE, 0);
	/* cs port */
	last_serial_minor_used++;
	hso_serial_start(odev, last_serial_minor_used, CS_INTERFACE, 0);

	// Does this REALLY do anything???
	usb_inc_dev_use( usb );

	/* Arm and submit the interrupt URB */
	usb_fill_int_urb(	odev->mux_intr_urb,
				odev->usb,
				usb_rcvintpipe(odev->usb, odev->mux_ep_intr),
				odev->mux_intr_buf,
				MUX_INTR_BUF_SIZE,
				intr_callback,
				odev,
				(odev->usb->speed == USB_SPEED_HIGH) ? ( 1 << odev->mux_bInterval) : odev->mux_bInterval );

	if ( (res = usb_submit_urb(odev->mux_intr_urb)) ) {
		WARN("%s failed mux_intr_urb %d", __FUNCTION__, res );
		goto exit;
	}

	// Okay, we are finally done...
	return odev;

exit:
	hso_free_memory(&odev);
	return NULL;
}

//=============================================================================
//	function:	hso_disconnect
//	purpose:	device removed, cleaning up
//=============================================================================
static void hso_disconnect( struct usb_device *usb, void *ptr )
{
	hso_priv *odev = ptr;
	struct usb_interface *intf = NULL;
	struct usb_config_descriptor *conf = NULL;
	int i = 0;

	D1("------------------------------------");

	// Sanity check!!!
	if ( !odev || !odev->usb ) {
		// We failed.  We are insane!!!
		WARN("unregistering non-existant device");
		return;
	}
	hso_serial_disconnect(usb, odev);

	for ( i=0 ; i < MUX_BULK_RX_BUF_COUNT; i++)
	{
		usb_unlink_urb( odev->mux_bulk_rx_urb_pool[i] );
	}
	usb_unlink_urb( odev->mux_bulk_tx_urb );
	usb_unlink_urb( odev->mux_intr_urb );

	// Make sure we fail the sanity check if we try this again.
	odev->usb = NULL;

	// It is possible that this function is called before
	// the "close" function.
	// This tells the close function we are already disconnected
	set_bit(HSO_NET_UNPLUG, &odev->flags);

	// We don't need the network device any more
	unregister_netdev( odev->net );

	// For sanity checks
	odev->net = NULL;

	// I ask again, does this do anything???
	usb_dec_dev_use( usb );

	conf = &( usb->config[odev->configuration_num] );

	// We are done with this interface, do some sanity checks first
	if (odev->interfaces[MUX_INTERFACE] != NULL && (intf = &( conf->interface[MUX_INTERFACE] )) != NULL && usb_interface_claimed( intf ) ) {
		usb_driver_release_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[MUX_INTERFACE]));
	}

	// We are done with this interface, do some sanity checks first
	if (odev->interfaces[QXDM_INTERFACE] != NULL && (intf = &( conf->interface[QXDM_INTERFACE] )) != NULL && usb_interface_claimed( intf ) ) {
		usb_driver_release_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[QXDM_INTERFACE]));
	}

	// We are done with this interface, do some sanity checks first
	if (odev->interfaces[CS_INTERFACE] != NULL && (intf = &( conf->interface[CS_INTERFACE] )) != NULL && usb_interface_claimed( intf ) ) {
		usb_driver_release_interface( &hso_driver, &(usb->config[odev->configuration_num].interface[CS_INTERFACE]));
	}

	// No more tied up kernel memory
	hso_free_memory(&odev);
}


//*****************************************************************************
//*****************************************************************************
// Helper functions
//*****************************************************************************
//*****************************************************************************

//=============================================================================
//	function:	check_and_claim_interfaces
//	purpose:	claim the interfaces 1 & 2, called from the probe
//			function of interface 0
//=============================================================================
static int check_for_claimed_interfaces( struct usb_config_descriptor *config, hso_priv *odev )
{
	struct usb_interface *intf_group = 0;

	if (odev->interfaces[CS_INTERFACE] == NULL || odev->interfaces[QXDM_INTERFACE] == NULL || odev->interfaces[MUX_INTERFACE] == NULL) {
		// Check claimed interfaces at next probe
		return 0;
	}

	intf_group = &( config->interface[MUX_INTERFACE] );
	if ( usb_interface_claimed( intf_group ) ) {
		D1("Interface 1 already claimed");
		return -1;
	}
	intf_group = &( config->interface[QXDM_INTERFACE] );
	if ( usb_interface_claimed( intf_group ) ) {
		D1("Interface 2 already claimed");
		return -1;
	}
	intf_group = &( config->interface[CS_INTERFACE] );
	if ( usb_interface_claimed( intf_group ) ) {
		D1("Interface 3 already claimed");
		return -1;
	}

	D1("No one has claimed our interface, good");
	return 0;
}

//=============================================================================
//	function:	get_mux_endpoints
//	purpose:	check the endpoints, and remember them if correct
//=============================================================================
static int get_mux_endpoints( struct usb_interface_descriptor *desc, hso_priv *odev)
{
	int i = 0;
	struct usb_endpoint_descriptor *endp = NULL;

	/* Start out assuming we won't find anything we can use */
	odev->mux_ep_bulk_out = odev->mux_ep_bulk_in = odev->mux_ep_intr = 0;

	for(i=0;i< desc->bNumEndpoints; i++) {
		endp = &(desc->endpoint[i]);

		if ( endp->bmAttributes != USB_ENDPOINT_XFER_BULK ) {
			if (endp->extralen) {
				if (parse_option_information( endp->extra, endp->extralen, odev )) {
					ERR("Could not get option specific data, not our device then");
					return -1;
				}

				odev->mux_ep_intr = (endp->bEndpointAddress & 0x7F);
				odev->mux_bInterval = endp->bInterval;
				odev->mux_ep_intr_size = endp->wMaxPacketSize;
			}
		} else {

			if ( endp->bEndpointAddress & USB_DIR_IN ) {
				odev->mux_ep_bulk_in = endp->bEndpointAddress & 0x7F;
				odev->mux_ep_bulk_in_size = endp->wMaxPacketSize;
			} else {
				odev->mux_ep_bulk_out = endp->bEndpointAddress;
				odev->mux_ep_bulk_out_size = endp->wMaxPacketSize;
			}
		}
	}

	/* Now make sure we got both an IN and an OUT */
	if (odev->mux_ep_bulk_in && odev->mux_ep_bulk_out && odev->mux_ep_intr) {
		D1( "detected BULK IN/OUT/interrupt packets of [size: %d/%d/%d, addr: %02x/%02x/%02x]", 
			odev->mux_ep_bulk_out_size, odev->mux_ep_bulk_out_size, odev->mux_ep_intr_size,
			odev->mux_ep_bulk_in,odev->mux_ep_bulk_out, odev->mux_ep_intr);
		return 0;
	}
	return -1;
}

//=============================================================================
//	function:	hso_serial_init
//	purpose:	initialize the serial connections, make the ttys
//=============================================================================
static int hso_serial_init()
{
	int i = 0;
	D1("Starting hso_serial");

	for(i = 0; i < HSO_SERIAL_TTY_MINORS; i++) {
		hso_serial_table[i] = NULL;
	}
	last_serial_minor_used = -1;

	/* register the tty driver */
	hso_serial_tty_driver.init_termios = tty_std_termios;
	hso_serial_tty_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;

	if (tty_register_driver(&hso_serial_tty_driver)) {
		D1("failed to register tty driver");
		return -1;
	}

	return 0;
}

//=============================================================================
//	function:	hso_serial_exit
//	purpose:	closes and cleans up serial interface
//=============================================================================
static void hso_serial_exit()
{
	hso_serial_t *serial = NULL;
	D1("Stopping hso_serial");
	int i = 0;

	for(i = 0; i < HSO_SERIAL_TTY_MINORS; i++) {
		serial = hso_serial_table[i];
		if(serial != NULL) {
			spin_lock_bh( &serial->serial_lock );
			while(serial->open_count > 0) {
				__hso_serial_close(serial);
			}
			D1("refcount %d", *hso_serial_tty_driver.refcount);
			hso_serial_stop(serial->odev->usb, i, serial->odev);
			spin_unlock_bh( &serial->serial_lock );
		}
	}

	D1(" unregistering %d", tty_unregister_driver(&hso_serial_tty_driver));
}

//=============================================================================
//	function:	hso_serial_start
//	purpose:	start the actual serial port
//=============================================================================
static hso_serial_t *hso_serial_start(hso_priv *odev, unsigned char minor, int type, unsigned char num)
{
	hso_serial_t *serial = NULL;

	D1("Registering serial %d", minor);
	if(minor >= HSO_SERIAL_TTY_MINORS) {
		ERR("%s %d minor out of range %d", __FUNCTION__, __LINE__, minor);
		return NULL;
	}

	if(!(serial = kmalloc(sizeof (hso_serial_t), GFP_KERNEL))) {
		ERR("%s - Out of memory", __FUNCTION__);
		return NULL;
	}

	spin_lock_init( &serial->serial_lock );
	serial->odev    = odev;
	serial->minor   = minor;
	serial->type	= type;
	serial->magic   = HSO_SERIAL_MAGIC;

	serial->rx_urb          = NULL;
	serial->rx_data         = NULL;
	serial->rx_data_length  = 0;

	/* get the endpoints if not muxed */
	if(type != MUX_INTERFACE) {
		// Get information on bulk endpoints
		if (get_std_serial_endpoints(serial, type, odev)) {
			ERR("Could not get bulk specific data, not our device then");
			goto exit;
		}
	}

	if (!(serial->rx_urb = usb_alloc_urb(0))) {
		ERR("Could not allocate urb?");
		goto exit;
	}
	serial->rx_urb->transfer_buffer = NULL;
	serial->rx_urb->transfer_buffer_length = 0;

	serial->rx_data_length  = (type == MUX_INTERFACE ? CTRL_URB_RX_SIZE : BULK_URB_RX_SIZE);
	if (!(serial->rx_data = kmalloc(serial->rx_data_length , GFP_KERNEL))) {
		ERR("%s - Out of memory", __FUNCTION__);
		goto exit;
	}
	memset(serial->rx_data, 0, serial->rx_data_length);

	serial->tx_data_length  = 0;
	serial->tx_data_count   = 0;
	serial->tx_buffer_count = 0;
	serial->tx_data         = NULL;
	serial->tx_urb          = NULL;

	if (!(serial->tx_urb = usb_alloc_urb(0))) {
		ERR("Could not allocate urb?");
		goto exit;
	}
	serial->tx_urb->transfer_buffer = NULL;
	serial->tx_urb->transfer_buffer_length = 0;

	serial->tx_data_length  = (type == MUX_INTERFACE ? CTRL_URB_TX_SIZE : BULK_URB_TX_SIZE);
	if (!(serial->tx_data = kmalloc( serial->tx_data_length, GFP_KERNEL))) {
		ERR("%s - Out of memory", __FUNCTION__);
		goto exit;
	}
	memset(serial->tx_data, 0, serial->tx_data_length);
	if (!(serial->tx_buffer = kmalloc(serial->tx_data_length, GFP_KERNEL))) {
		ERR("%s - Out of memory", __FUNCTION__);
		goto exit;
	}
	memset(serial->tx_buffer, 0, serial->tx_data_length);

	/* The muxed ports can not use full 64 bytes and the diagnostic must use the full 64 bytes if sending larger packets. */
	if(type == MUX_INTERFACE) {
		serial->tx_data_length --;
	}

	if(type == MUX_INTERFACE) {
		serial->write_data = hso_mux_serial_write_data;
	} else {
		serial->write_data = hso_std_serial_write_data;
	}

	serial->open_count = 0;
	serial->private = NULL;

	serial->tqueue.routine = port_softint;
	serial->tqueue.data = serial;

	serial->read_scheduled = 0;

	hso_serial_table[minor] = serial;

	/* notify hotplug */
	if (hotplug) {
		call_policy_interface("add", serial->minor, odev->usb->descriptor.idVendor, odev->usb->descriptor.idProduct);
	}

	return serial;

exit:
	if (serial->rx_urb) {
		usb_free_urb(serial->rx_urb);
	}
	if (serial->rx_data) {
		kfree(serial->rx_data);
	}
	if (serial->tx_data) {
		kfree(serial->tx_data);
	}
	if (serial->tx_urb) {
		usb_free_urb(serial->tx_urb);
	}
	
	if (serial) {
		kfree(serial);
	}
	return NULL;
}

//=============================================================================
//	function:	get_std_serial_endpoints
//	purpose:	check and remember the endpoints for the diag and CS
//=============================================================================
static int get_std_serial_endpoints(hso_serial_t *serial, int ifacenr, hso_priv *odev)
{
	struct usb_interface_descriptor *desc = &(odev->interfaces[ifacenr]->altsetting[odev->interfaces[ifacenr]->act_altsetting]);
	int i = 0;
	struct usb_endpoint_descriptor *endp = NULL;

	/* Start out assuming we won't find anything we can use */
	serial->ep_bulk_out = 0;
	serial->ep_bulk_in = 0;

	for(i=0;i< desc->bNumEndpoints; i++) {
		endp = &(desc->endpoint[i]);

		if ( endp->bmAttributes != USB_ENDPOINT_XFER_BULK ) {
			continue;
		}

		/* Check the first endpoint to see if it is IN or OUT */
		if ( endp->bEndpointAddress & USB_DIR_IN ) {
			serial->ep_bulk_in = endp->bEndpointAddress & 0x7F;
			serial->ep_bulk_in_size = endp->wMaxPacketSize;
		} else {
			serial->ep_bulk_out = endp->bEndpointAddress;
			serial->ep_bulk_out_size = endp->wMaxPacketSize;
		}
	}

	/* Now make sure we got both an IN and an OUT */
	if (serial->ep_bulk_in && serial->ep_bulk_out) {
		D1( "detected BULK IN/OUT packets of [size: %d/%d, addr: %02x/%02x]", 
			serial->ep_bulk_in_size, serial->ep_bulk_out_size,
			serial->ep_bulk_in,serial->ep_bulk_out);
		return 0;
	}
	return -1;
}

//=============================================================================
//	function:	hso_serial_disconnect
//	purpose:	Called once per device shutdown, usb unused
//=============================================================================
static void hso_serial_disconnect(struct usb_device *usb, hso_priv *odev)
{
	hso_serial_t *serial = NULL;
	int i = 0;
	
	for(i = 0; i < HSO_SERIAL_TTY_MINORS; i++)
	{
		serial = hso_serial_table[i];
		if(serial != NULL)
		{
			spin_lock_bh( &serial->serial_lock );
			while(serial->open_count > 0)
			{
				__hso_serial_close(serial);
			}
			hso_serial_stop(usb, i, odev);
			spin_unlock_bh( &serial->serial_lock );
		}
	}
	last_serial_minor_used=-1;
}

//=============================================================================
//	function:	__hso_serial_close
//	purpose:	decrement count and unlink URBS if no more connected. 
//			Calling function should hold serial_lock if needed.
//=============================================================================
static void __hso_serial_close(hso_serial_t *serial)
{
	if (!serial->open_count) {
		err ("%s - port %d: not open", __FUNCTION__, serial->minor);
		return;
	}

	D("Port %d ,open count: %d", serial->minor, serial->open_count);
	--serial->open_count;

	if(serial->open_count <= 0) {
		if( serial->type != MUX_INTERFACE)
		{
			D1("Unlink any pending bulk URBS");
			/* Stop pending urbs. Do not deallocate them, it will be done in hso_serial_stop. */
			/* Is this in interrupt state? */
			usb_unlink_urb(serial->rx_urb);
			usb_unlink_urb(serial->tx_urb);
		}
		/*close ? */
		serial->open_count = 0;
		clear_bit(HSO_PRIV_FLAG_TX_SENT, &serial->flags);
		clear_bit(HSO_PRIV_FLAG_RX_SENT, &serial->flags);
		if(serial->tty) {
			serial->tty->driver_data = NULL;
			serial->tty = NULL;
		}
	}
}

//=============================================================================
//	function:	hso_serial_stop
//	purpose:	release the serial port
//=============================================================================
static void hso_serial_stop(struct usb_device *dev, unsigned char minor, hso_priv *odev)
{
	hso_serial_t *serial;

	D1("Deregistering serial %d", minor);
	if(minor >= HSO_SERIAL_TTY_MINORS) {
		ERR("%s %d minor out of range %d", __FUNCTION__, __LINE__, minor);
		return;
	}
	serial = hso_serial_table[minor];
	if(serial == NULL || serial->magic != HSO_SERIAL_MAGIC) {
		ERR("Trying to deregister an unused serial device");
		return;
	}

	/* notify hotplug */
	if (hotplug) {
		call_policy_interface("remove", serial->minor, odev->usb->descriptor.idVendor, odev->usb->descriptor.idProduct);
	}

	serial->odev = NULL;
	serial->minor = 0;
	serial->magic = 0;
	serial->write_data = NULL;

	if(serial->rx_urb != NULL) {
		usb_unlink_urb(serial->rx_urb);
		usb_free_urb(serial->rx_urb);
	}
	serial->rx_urb = NULL;

	if(serial->rx_data != NULL) {
		kfree(serial->rx_data);
	}
	serial->rx_data = NULL;

	if(serial->tx_urb != NULL) {
		usb_unlink_urb(serial->tx_urb);
		usb_free_urb(serial->tx_urb);
	}
	serial->tx_urb = NULL;

	if(serial->tx_data != NULL) {
		kfree(serial->tx_data);
	}
	serial->tx_data = NULL;

	kfree(serial);
	serial = NULL;
	hso_serial_table[minor] = NULL;
}

//=============================================================================
//	function:	hso_alloc_memory
//	purpose:	allocate all memory needed in our device
//=============================================================================
static int hso_alloc_memory(hso_priv **_odev)
{
	hso_priv *odev = *_odev;
	int i = 0;

	if(!(odev = kmalloc(sizeof(hso_priv), GFP_KERNEL))) {
		ERR("out of memory allocating device structure");
		return -1;
	}

	// Zero everything out.
	memset(odev, 0, sizeof(hso_priv));

	if (!(odev->mux_intr_urb = usb_alloc_urb(0))) {
		ERR("Could not allocate urb?");
		return -1;
	}

	if (!(odev->mux_intr_buf = kmalloc(MUX_INTR_BUF_SIZE, GFP_KERNEL))) {
		ERR("Could not allocate buf?");
		return -1;
	}
	memset(odev->mux_intr_buf, 0, MUX_INTR_BUF_SIZE);

	for( i = 0; i < MUX_BULK_RX_BUF_COUNT ; i++) {
		if (!(odev->mux_bulk_rx_urb_pool[i] = usb_alloc_urb(0))) {
			ERR("Could not allocate urb?");
			return -1;
		}
		if (!(odev->mux_bulk_rx_buf_pool[i] = kmalloc(MUX_BULK_RX_BUF_SIZE, GFP_KERNEL))) {
			ERR("Could not allocate buf?");
			return -1;
		}
		memset(odev->mux_bulk_rx_buf_pool[i], 0, MUX_BULK_RX_BUF_SIZE);
	}

	if (!(odev->mux_bulk_tx_urb = usb_alloc_urb(0))) {
		ERR("Could not allocate urb?");
		return -1;
	}
	if (!(odev->mux_bulk_tx_buf = kmalloc(MUX_BULK_TX_BUF_SIZE, GFP_KERNEL))) {
		ERR("Could not allocate buf?");
		return -1;
	}
	memset(odev->mux_bulk_tx_buf, 0, MUX_BULK_TX_BUF_SIZE);

	*_odev = odev;
	return 0;
}

//=============================================================================
//	function:	hso_free_memory
//	purpose:	free all memory used in our device
//=============================================================================
static void hso_free_memory(hso_priv **_odev)
{
	hso_priv *odev = *_odev;
	int i = 0;

	if (odev) {
		if (odev->mux_intr_urb) {
			usb_free_urb(odev->mux_intr_urb);
		}
		if (odev->mux_intr_buf) {
			kfree(odev->mux_intr_buf);
		}
		for(i = 0; i < MUX_BULK_RX_BUF_COUNT; i++) {
			if (odev->mux_bulk_rx_urb_pool[i]) {
				usb_free_urb(odev->mux_bulk_rx_urb_pool[i]);
			}
			if (odev->mux_bulk_rx_buf_pool[i]) {
				kfree(odev->mux_bulk_rx_buf_pool[i]);
			}
		}
		if (odev->mux_bulk_tx_urb) {
			usb_free_urb(odev->mux_bulk_tx_urb);
		}
		if (odev->mux_bulk_tx_buf) {
			kfree(odev->mux_bulk_tx_buf);
		}

		kfree(odev);
		odev = NULL;
	}

	*_odev = NULL;
}

//=============================================================================
//	function:	set_ethernet_addr
//	purpose:	setting the mac-address of the newly created iface
//=============================================================================
static void set_ethernet_addr( hso_priv *odev )
{
	unsigned char mac_addr[6];
	int i = 0;
	int len = 0;
	unsigned char buffer[13];

	/* Let's assume we don't get anything */
	mac_addr[0] = 0x00;
	mac_addr[1] = 0x00;
	mac_addr[2] = 0x00;
	mac_addr[3] = 0x00;
	mac_addr[4] = 0x00;
	mac_addr[5] = 0x00;

	if (0 > (len = usb_string(odev->usb, odev->usb->descriptor.iSerialNumber, buffer, 13))) {
		ERR("Attempting to get MAC address failed: %d", -1*len);
		return;
	}

	/* Sanity check */
	if (len != 12)	{
		/* You gotta love failing sanity checks */
		ERR("Attempting to get MAC address returned %d bytes", len);
		return;
	}

	/* Fill in the mac_addr */
	for (i = 0; i < 6; i++) {
		if ((16 == buffer[2 * i]) || (16 == buffer[2 * i + 1])) {
			ERR("Bad value in MAC address i:%d", i);
		} else {
			mac_addr[i] = ( hex2dec( buffer[2 * i] ) << 4 ) + hex2dec( buffer[2 * i + 1] );
		}
	}

	mac_addr[0] = 0x00;
	mac_addr[1] = 0x03;

	/* Now copy it over to our network device structure */
	memcpy( odev->net->dev_addr, mac_addr, sizeof(mac_addr) );

	/* Create the default fake ethernet header. */
	memcpy(odev->dummy_eth_head.h_dest, mac_addr , ETH_ALEN);
	memcpy(odev->dummy_eth_head.h_source, dummy_mac, ETH_ALEN);
	odev->dummy_eth_head.h_proto = htons(ETH_P_IP);
}

//=============================================================================
//	function:	log_device_info
//	purpose:	log all device info to syslog
//=============================================================================
void log_device_info(hso_priv *odev)
{
	unsigned char manu[256];
	unsigned char prod[256];
	unsigned char sern[256];
	unsigned char *mac_addr;

	/* Try to get the device Manufacturer */
	get_string(manu, sizeof(manu), odev->usb->descriptor.iManufacturer, odev);

	/* Try to get the device Product Name */
	get_string(prod, sizeof(prod), odev->usb->descriptor.iProduct, odev);

	/* Try to get the device Serial Number */
	get_string(sern, sizeof(sern), odev->usb->descriptor.iSerialNumber, odev);

	/* This makes it easier for us to print */
	mac_addr = odev->net->dev_addr;

	/* Now send everything we found to the syslog */
	NFO( "'%s': %s %s %s", odev->net->name, manu, prod, sern);
	NFO( "%s: %02X:%02X:%02X:%02X:%02X:%02X", odev->net->name, mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5] );
}

//=============================================================================
//	function:	parse_option_information
//	purpose:	
//=============================================================================
static int parse_option_information( unsigned char *data, int length, hso_priv *odev )
{
	int i = 0;

	if (length != 3) {
		ERR("Extralen don't contain expected data? length:%d", length);
		return -1;
	}

	/* Length */
	odev->option_info.length = data[i++];

	/* Type */
	odev->option_info.descriptor_type = data[i++];

	if (odev->option_info.descriptor_type != CS_INTERFACE_DESC) {
		ERR("Expected CS_INTERFACE_DESC, got: %02x", odev->option_info.descriptor_type);
		return-1;
	}

	/* Enabled modem ports */
	odev->option_info.enabled_ports = data[i++];

	return 0;
}

//=============================================================================
//	function:	find_valid_configuration
//	purpose:	
//=============================================================================
static int find_valid_configuration( struct usb_device *device, hso_priv *odev )
{
	struct usb_config_descriptor *conf = NULL;
	int conf_num = 0;

	/* We expect device to have only one configuration */
	if (device->descriptor.bNumConfigurations != 1) {
		WARN("Assuming the first configuration should be used");
	}

	// Use the first configuration.
	conf_num = 0;
	conf = &( device->config[conf_num] );

	// 0. usable endpoint - hsdpa net + muxed serial ports.
	// 1. debug port. (QXDM port)
	// 2. Modem port. unused - only usable in Circuit switched connections.
	if ( conf->bNumInterfaces > 3 ) {
		ERR("We don't see <=3 interfaces? (%d)", conf->bNumInterfaces);
		// I currently don't know how to handle devices with any number of interfaces
		// other than 2.
		return -1;
	}

	// As we are going to use the first configuration, save the configurationValue so we know what to refer to
	odev->configuration_num = conf_num;
	odev->bConfigurationValue = conf->bConfigurationValue;

	// Check if we found an option device
	if (check_interface(device, odev, MUX_INTERFACE)) {
		D1("No configuration suited us");
		return -1;
	}

	// Check if we find qxdm interface
	if (check_interface(device, odev, QXDM_INTERFACE)) {
		D1("No configuration suited us");
		return -1;
	}

	// Check if we find cs interface
	if (check_interface(device, odev, CS_INTERFACE)) {
		D1("No configuration suited us");
		return -1;
	}

	// We found one.
	return 0;
}

//=============================================================================
//	function:	check_interface
//	purpose:	
//=============================================================================
static int check_interface( struct usb_device *device, hso_priv *odev, type_intf type )
{
	struct usb_config_descriptor *conf = NULL;
	struct usb_interface *intf = NULL;
	struct usb_interface_descriptor *alts = NULL;

	if (!(conf = &( device->config[odev->configuration_num] ))) {
		ERR("Could not get configuration?");
		return -1;
	}

	if (!(intf = usb_ifnum_to_if(device, type))) {
		ERR("Could not get interface?");
		return -1;
	}

	if (intf->num_altsetting != 1) {
		WARN("More alternatives, assume to use active one (%d/%d)", intf->act_altsetting, intf->num_altsetting);
	}

	// There should be only one alternative
	if (!(alts = &(intf->altsetting[intf->act_altsetting]))) {
		ERR("Could not get interface descriptor?");
		return -1;
	}

	odev->interfaces[type] = intf;
	if(type == MUX_INTERFACE) {
		// Get information on bulk endpoints
		if (get_mux_endpoints(alts, odev)) {
			ERR("Could not get bulk specific data, not our device then");
			return -1;
		}
		D("modem interface = %d %d" , alts->bInterfaceNumber, intf->act_altsetting);

		// Save the interface number so we know what to enable later
		odev->mux_bInterfaceNumber = alts->bInterfaceNumber;
		odev->mux_altsetting = intf->act_altsetting;
	};
	return 0;
}

//=============================================================================
//	function:	hex2dec
//	purpose:	
//=============================================================================
static inline unsigned char hex2dec( unsigned char digit )
{
	if ( (digit >= '0') && (digit <= '9') ) {
		return (digit - '0');
	}

	// Map all characters to 0-15
	if ( (digit >= 'a') && (digit <= 'z') ) {
		return (digit - 'a' + 10) % 16;
	}
	if ( (digit >= 'A') && (digit <= 'Z') ) {
		return (digit - 'A' + 10) % 16;
	}

	return 16;
}

//=============================================================================
//	function:	get_string
//	purpose:	
//=============================================================================
static void get_string(u8 *buf, int buf_len, int string_num, hso_priv *odev)
{
	int len = 0;

	if (!buf) {
		ERR("No buffer?");
		return;
	}

	buf[0]= 0x00;
	if (string_num) {
		// Put it into its buffer
		len = usb_string(odev->usb, string_num, buf, buf_len);
		// Just to be safe
		buf[len] = 0x00;
	}
}

//=============================================================================
//	function:	call_policy_interface
//	purpose:	
//=============================================================================
static void call_policy_interface (char *verb, int line, __u16 idVendor, __u16 idProduct)
{
	char *argv [3];
	char **envp = NULL;
	char *buf = NULL;
	char *scratch = NULL;
	int i = 0;
	int value = 0;
	int scratch_size = 256;

	if (!hotplug_path [0]) {
		return;
	}

	if (in_interrupt ()) {
		D1("in_interrupt");
		return;
	}

	if (!(envp = (char **) kmalloc (20 * sizeof (char *), GFP_KERNEL))) {
		return;
	}
	if (!(buf = kmalloc (scratch_size, GFP_KERNEL))) {
		kfree (envp);
		return;
	}

	/* only one standardized param to hotplug command: type */
	argv [0] = hotplug_path;
	argv [1] = "modem";
	argv [2] = 0;

	/* minimal command environment */
	envp [i++] = "HOME=/";
	envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

#ifdef	DEBUG
	/* hint that policy agent should enter no-stdout debug mode */
	envp [i++] = "DEBUG=kernel";
#endif
	/* extensible set of named bus-specific parameters, supporting multiple driver selection algorithms. */
	scratch = buf;

	/* action:  add, remove */
	envp [i++] = scratch;
	scratch += sprintf (scratch, "ACTION=%s", verb) + 1;

	envp [i++] = scratch;
	scratch += sprintf (scratch, "DEVICE=/dev/ttyHS%d", line) + 1;

	/* class-based driver binding models */
	envp [i++] = scratch;
	scratch += sprintf (scratch, "TYPE=modem")+1;

	/* Adding vendor  */
	envp [i++] = scratch;
	scratch += sprintf (scratch, "VENDOR=0x%x", idVendor)+1;
	
	/* Adding product  */
	envp [i++] = scratch;
	scratch += sprintf (scratch, "PRODUCT=0x%x", idProduct)+1;

	envp [i++] = 0;

	/* NOTE: user mode daemons can call the agents too */
	D2("%s %s", argv [0], verb);
	value = call_usermodehelper (argv [0], argv, envp);
	kfree (buf);
	kfree (envp);
	if (value != 0) {
		D2("usbserial policy returned 0x%x", value);
	}
}

//=============================================================================
//	function:	get_serial
//	purpose:	returns serial struct and checks for validity of tty struct.
//=============================================================================
static inline hso_serial_t *get_serial(struct tty_struct *tty)
{

	if(tty == NULL || tty->driver_data == NULL) {
		return NULL;
	}

	return (hso_serial_t *)tty->driver_data;
}

//=============================================================================
//	function:	hso_serial_rx_flip
//	purpose:	
//=============================================================================
static void hso_serial_rx_flip(void *priv)
{
	struct tty_struct *tty = (struct tty_struct *)priv;
	hso_serial_t *serial = get_serial(tty);
	unsigned char data[HSO_RX_MAX_FLIP_SIZE];
	unsigned char fp[HSO_RX_MAX_FLIP_SIZE];
	int len = 0;
	int room = 0;

	D1(" ");

	if((tty == NULL) || (serial == NULL)) {
		return;
	}

	D1("room =  %d ", tty->ldisc.receive_room(tty));

	if ((room = tty->ldisc.receive_room(tty))) {
		D1(" ");
		if ((len = get_rxbuf_data(data, min(room, HSO_RX_MAX_FLIP_SIZE), serial))) {
			D1("port %d - %d bytes (room %d, max %d) bytes in buffer %d", serial->minor, len, room, HSO_RX_MAX_FLIP_SIZE, (&serial->rx_fifo)->cnt);
			memset(fp, 0, sizeof(fp));	/* Could be len instead of sizeof. */
			tty->ldisc.receive_buf(tty, data, fp, len);
		}
	}
	D1(" ");

	if (len || !room) {
		schedule_task(&serial->rx_flip_task); /* Reschedule it. */
	}
}

//=============================================================================
//	function:	get_rxbuf_data
//	purpose:	
//=============================================================================
static int get_rxbuf_data(unsigned char *data, unsigned short len, hso_serial_t *serial)
{
	unsigned char *p_b = data;
	fifo_t *p_f = NULL;
	int n = len;
	int m = 0;
	unsigned int flags = 0;
	int endlen = 0;

	D1("%s - port %d", __FUNCTION__, serial->minor);
	p_f = &serial->rx_fifo;

	/* disable irq ... */
	save_flags(flags);
	cli();

	/* check if empty ... */
	if (p_f->p_h == p_f->p_t) {
		/* enable irq ... */
		restore_flags(flags);
		return 0;
	}

	/* calc ... */
	n = m = min((int)len, (int) (p_f->cnt) );

	/* if n is less than endlen one chunk else split it up in two. */
	endlen = &p_f->p_b[(p_f->size)-1] - p_f->p_t + 1 ;

	if ( n < endlen ) {
		memcpy(p_b, p_f->p_t, n );
		p_f->p_t +=n;
	} else {
		memcpy(p_b, p_f->p_t , endlen);
		memcpy(p_b + endlen, p_f->p_b , n - endlen);
		p_f->p_t = p_f->p_b + (n - endlen);
	}
	p_f->cnt -=n;
	if ( p_f->cnt <= HSO_THRESHOLD_UNTHROTTLE && test_and_clear_bit(HSO_PRIV_FLAG_THROTTLED, &serial->flags) ) {
		D1("Unthrottle: p_f->cnt[%d] < HSO_THRESHOLD_UNTHROTTLE[%d]", p_f->cnt, HSO_THRESHOLD_UNTHROTTLE);
	}

	/* enable irq ... */
	restore_flags(flags);

	/* return cnt read ... */
	return m;
}

//=============================================================================
//	function:	port_softint
//	purpose:	
//=============================================================================
static void port_softint(void *private)
{
	hso_serial_t *serial = (hso_serial_t *)private;
	struct tty_struct *tty;

	D1(" - port %d", serial->minor);
	
	if (!serial) {
		return;
	}

	tty = serial->tty;
	if (!tty) {
		return;
	}

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) {
		D1(" - write wakeup call.");
		(tty->ldisc.write_wakeup)(tty);
	}

	wake_up_interruptible(&tty->write_wait);
}


//*****************************************************************************
//*****************************************************************************
// Module definitions
//*****************************************************************************
//*****************************************************************************
module_init( hso_init );
module_exit( hso_exit );

MODULE_AUTHOR(MOD_AUTHOR);
MODULE_DESCRIPTION(MOD_DESCRIPTION);
MODULE_LICENSE(MOD_LICENSE);
MODULE_DEVICE_TABLE(usb, hso_ids);

MODULE_PARM_DESC(debug, "Level of debug [0x01 | 0x02 | 0x04 | 0x08]");
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(hotplug, "Will enable-1/disable-0 hotplug calls to user space");
MODULE_PARM(hotplug, "i");

